import core.helper as helper
import review.spellcheck as s_check

import re
import os


def spellcheck_string(text, playlist_name):
    spellcheck_dict = s_check.main_dict
    for dict_s in spellcheck_dict:
        if dict_s == 'regular':
            for x, y in spellcheck_dict[dict_s].items():
                try:
                    text = re.sub(x, y, text, flags=re.IGNORECASE)
                except:
                    text = text.replace(x, y)
        elif dict_s == playlist_name:
            for x, y in spellcheck_dict[dict_s].items():
                try:
                    text = re.sub(x, y, text, flags=re.IGNORECASE)
                except:
                    text = text.replace(x, y)
    return text


def spellcheck(input_path, playlist_name, jsons):
    edited_path = input_path + helper.EDITFOLDER + '\\'
    if not os.path.exists(edited_path):
        os.makedirs(edited_path)
    with open(edited_path + "_text.txt", "w", encoding='utf8') as f:
        for filename in jsons:
            transcript = helper.get_transcript(input_path, filename)
            # temp = transcript['text']
            found = re.search(r'[^,.!\?]{2000}', transcript['text'])
            if found is not None:
                print('stuck in no-punctuation mode: ' + filename)
            # https://github.com/openai/whisper/discussions/194
            # TODO: redo translation or sth
            # Status: Yet to be fixed
            transcript['text'] = spellcheck_string(
                transcript['text'], playlist_name)
            # TODO: Improve segment checks if correction happens in text, but is split in segment:
            # Port Pala -> "... Pot", "Pala ..."
            for segment in transcript['segments']:
                segment['text'] = spellcheck_string(
                    segment['text'], playlist_name)
            # TODO: make a check if no need to write to file again
            f.write(transcript['text'] + '\n')

            if not os.path.exists(edited_path):
                os.makedirs(edited_path)

            helper.dict2json(transcript, filename, edited_path)

    # do_token_stuff(input_path, jsons)
    # Todo: make every subsequential call look for transcripts in edited


def add_transcript(input_path, playlist_name, playlist_info, episodes_info):
    # Phase 0: Setup
    audiofolder = helper.AUDIOFOLDER

    data_files, data_folders = helper.extract_file_folder(input_path)
    data_files = helper.show_newest_files(input_path, data_files)
    # --- 1. Setup --- #
    # Log file
    # check what info is available

    # Read existing info
    playlist_info = {}
    episodes_info = {}

    # If major changes are made:
    playlist_info, episodes_info = helper.setup_infos(
        playlist_info, episodes_info, input_path)

    if not data_folders or audiofolder not in data_folders:
        return

    for foldername in data_folders:
        if foldername == audiofolder:
            data_files_audio, data_folders_audio = helper.extract_file_folder(
                input_path + foldername + '\\')

    jsons = []
    audios = []

    for filename in data_files_audio:
        split_tup = os.path.splitext(input_path + foldername + '\\' + filename)
        file_name = split_tup[0]
        file_extension = split_tup[1]
        # TODO: if filename.endswith('.json'):
        if file_extension == '.json':
            jsons.append(filename)
        elif file_extension == '.mp3':
            audios.append(filename)
            # jsons.append(filename.replace('.mp3', '.json'))
        else:
            print(filename)
    # phase 1: Addapt names
    phase_1 = False
    if phase_1:
        # TODO: generalize this
        for filename in jsons:
            entry = filename[:4]
            # if playlist_name not in entry:
            #     continue
            title = filename.replace('.json', '')
            try_this = helper.title_mine(title, playlist_name)
            for idx, eID in enumerate(episodes_info):
                new_title = episodes_info[eID]['title']
                for each in helper.noFileChars:
                    new_title = new_title.replace(each, '')
                comp = helper.title_mine(new_title, playlist_name)
                is_match = helper.compare_titles(try_this, comp, playlist_name)
                matched = False
                if is_match >= 1:
                    matched = True
                else:
                    matched = helper.similar(title, new_title, "levenshtein")
                if matched:
                    clean_name = helper.fill_digits(idx, 3) + '_' + new_title
                    path_old = input_path + audiofolder + '\\' + filename
                    path_new = input_path + audiofolder + '\\' + clean_name + '.json'
                    if os.path.exists(path_old):
                        if os.path.exists(path_new):
                            continue
                        os.rename(path_old, path_new)
                    filename = filename.replace('.json', '.mp3')
                    path_old = input_path + audiofolder + '\\' + filename
                    path_new = input_path + audiofolder + '\\' + clean_name + '.mp3'
                    if filename in audios:
                        if os.path.exists(path_old) and not os.path.exists(path_new):
                            if not os.path.exists(path_new):
                                os.rename(path_old, path_new)
    spellcheck(input_path, playlist_name, jsons)
    # ! TODO: make this way more permanent
    # TODO

    def rreplace(s, old, new, occurrence):
        li = s.rsplit(old, occurrence)
        return new.join(li)

    if len(episodes_info) < len(jsons):
        for file in jsons:
            eID = file.split("_", 1)[0]
            if eID not in episodes_info:
                title = file.split("_", 1)[1]
                title = rreplace(title, ".json", "", 1)
                episodes_info[eID] = {'title': title}
        playlist_info['title'] = playlist_name
        playlist_info['author_name'] = playlist_name
        helper.dict2json(playlist_info, "playlist_info", input_path)
        helper.dict2json(episodes_info, "episodes_info", input_path)


def main(data_pl_path, pl_n, playlist_info, episodes_info):
    print('add_transcript', flush=True)
    add_transcript(data_pl_path, pl_n, playlist_info, episodes_info)


if __name__ == '__main__':
    main()



import publish.graphvizify as graphvizify
import config.personal as personal
import core.helper as helper

from PIL import Image
from numpy.linalg import norm
import numpy as np
import json
import re
import os

# "path\to\your\Obsidian\vault"
OBSIDIAN_PATH = personal.OBSIDIAN_PATH


def add_links(links):
    note = "# Links\n"
    if links:
        link_preps = []
        for link in links.keys():
            link_preps.append("lexicon/" + link)
        note += "[[" + "]] ; [[".join(link_preps) + "]]"
    note += "\n\n"
    # text = re.sub(r"\b\w+\b", r"[[\g<0>]]", text)
    return note


def add_related(group_list, pl_n):
    note = "# Related Episodes\n"
    g_list = []
    for i in group_list:
        g_list.append(f"{pl_n}/{i}|{i}")
    note += "[[" + "]] ; [[".join(g_list) + "]]\n\n"
    # text = re.sub(r"\b\w+\b", r"[[\g<0>]]", text)
    return note


def add_urls_to_note(url, prefix):
    note = ""
    urls = url.split(" ; ")
    brands = []
    for url in urls:
        brand = helper.get_brand(url)
        if brand in brands:
            brand += str(brands.count(brand))
        brands.append(brand)
        note += prefix + "url_" + brand + ": \"" + url + "\"\n"
    return note


def make_a_note_out_of_me(transcript, episode_info, playlist_info):
    note = "---\n"
    tags = ""
    if len(transcript["text"]) < 50:
        tags += "#missingText "

    if "date" in episode_info:
        note += "date: " + episode_info["date"] + "\n"
    if "runtime" in episode_info:
        note += "runtime: " + episode_info["runtime"] + "\n"
    if "title" in episode_info:
        note += "title: \"" + episode_info["title"] + "\"\n"
    if "url" in episode_info:
        note += add_urls_to_note(episode_info["url"], "")
    if "playlist" in playlist_info:
        tags += "#" + playlist_info["playlist"] + " "
        note += "playlist: \"" + playlist_info["playlist"] + "\"\n"
    if "author_type" in playlist_info:
        note += "author_type: \"" + playlist_info["author_type"] + "\"\n"
    if "author_name" in playlist_info:
        tags += "#" + playlist_info["author_name"] + " "
        note += "author_name: \"" + playlist_info["author_name"] + "\"\n"
    if "author_url" in playlist_info:
        note += add_urls_to_note(playlist_info["author_url"], "author_")
    if "publisher" in playlist_info:
        note += "publisher: \"" + playlist_info["publisher"] + "\"\n"
    if "language" in playlist_info:
        tags += "#" + playlist_info["language"] + " "
        note += "language: \"" + playlist_info["language"] + "\"\n"
    if "@type" in playlist_info:
        tags += "#" + playlist_info["@type"] + " "
        note += "@type: \"" + playlist_info["@type"] + "\"\n"
    if "accessMode" in playlist_info:
        tags += "#" + playlist_info["accessMode"] + " "
        note += "accessMode: \"" + playlist_info["accessMode"] + "\"\n"
    if "url" in playlist_info:
        note += add_urls_to_note(playlist_info["url"], "playlist_")
    if "image" in playlist_info:
        note += "playlist_image: \"" + playlist_info["image"] + "\"\n"
    note += "---\n"
    # note += "```citeTitle:\n" +  + "\n```\n"
    # note += "```citeID:\n" + episode_info["citeID"] + "\n```\n"
    # note += "---\n"
    # note += "url: " + episode_info["url"] + "\n"
    # note += "zotero: " + episode_info["zotero"] + "\n"
    # note += "\n---\n"
    note += "tags:: " + tags + "\n"
    did_i_read_it = "#read/false "
    note += "read?:: " + did_i_read_it + "\n"
    comment = ""
    note += "comment:: " + comment + "\n"
    note += "\n---\n"
    return note


def add_full_text(text, links):
    for lemma, words in links.items():
        for word in words:
            text = re.sub(r"\b" + word + r"\b",
                          r"[[lexicon/" + lemma + "|\g<0>]]", text, flags=re.I)
    note = "# FullText\n"
    note += text + "\n\n"
    return note


def make_images(sim_matrix, edited_path):

    I8 = (((sim_matrix - sim_matrix.min()) /
          (sim_matrix.max() - sim_matrix.min())) * 255.9).astype(np.uint8)

    img = Image.fromarray(I8)
    factor = 9
    new_size = (img.size[0] * factor, img.size[1] * factor)
    img.save(edited_path + "_sim_matrix.png")
    img = img.resize(new_size, Image.Resampling.NEAREST)
    img.save(edited_path + f"_sim_matrix{factor}.png")

    I8 = I8 + I8.transpose()

    img = Image.fromarray(I8)
    factor = 9
    new_size = (img.size[0] * factor, img.size[1] * factor)
    img.save(edited_path + "_sim_matrixM.png")
    img = img.resize(new_size, Image.Resampling.NEAREST)
    img.save(edited_path + f"_sim_matrixM{factor}.png")

    np.savetxt(edited_path + "_sim_matrix_I8.csv", I8, delimiter=";")


def dotify(dict_var):
    # todo: Rework this
    episodes_info = {}
    lexicon_ep_count = {}
    linkpreps = {}
    ##
    dot = {}
    for eID, episode_info in episodes_info.items():
        words = {}
        if eID not in linkpreps:
            continue
        # if eID > 100:
        #     break
        for word, lemma in linkpreps[eID].items():
            count = 0
            if word in lexicon_ep_count:
                if eID in lexicon_ep_count[word]:
                    count = lexicon_ep_count[word][eID]
            if lemma not in words:
                words[lemma] = 0
            words[lemma] += count
        ep = {
            "title": episode_info["title"],
            "words": words
        }
        dot.update({eID: ep})


def include_episodes(edited_path, obs_pl_path, episodes_info, playlist_info):
    with open(edited_path + '_lexicon_ep_count.json', encoding='utf-8') as json_file:
        lexicon_ep_count = json.load(json_file)
    with open(edited_path + '_lemma_dict.json', encoding='utf-8') as json_file:
        lemma_dict = json.load(json_file)
    with open(edited_path + '_lemmacon.json', encoding='utf-8') as json_file:
        lemmacon = json.load(json_file)
    trivial_max = len(episodes_info) * 0.1
    trivial_min = len(episodes_info) * 0.01
    trivial_min_in_e = 3
    link_min = 2
    wordlen_min = 2
    groups = {}
    words_by_ep = {}
    e_vector = {}
    for lemma, words in lemmacon.items():
        count = sum(words.values())
        if count < trivial_min:
            # This lemma is insignificant
            continue
        episodes_with_word_and_count = {}
        for word in words:
            if len(word) < wordlen_min:
                # word is to short (here: 1 symbol)
                continue
            episodes = lexicon_ep_count[word]
            for eID, e_count in episodes.items():
                eID = int(eID)
                if e_count >= trivial_min_in_e:
                    if eID not in episodes_with_word_and_count:
                        episodes_with_word_and_count[eID] = {}
                    episodes_with_word_and_count[eID].update({word: e_count})
        if len(episodes_with_word_and_count) > trivial_max:
            # number of episodes where this word is significant is too high
            continue
        if len(episodes_with_word_and_count) < link_min:
            # number of episodes where this word is significant is too low
            continue
        else:
            for key, value in episodes_with_word_and_count.items():
                if key not in words_by_ep:
                    words_by_ep[key] = {}
                words_by_ep[key].update({lemma: value})
            e_vector[lemma] = 0

    words_by_ep = dict(sorted(words_by_ep.items()))

    # Now we have:
    # words_by_ep = {
    #   "0" : {
    #       "lemma1" : {
    #           "word1" : 5,
    #           "word2" : 17
    #       }
    #   }
    # }

    # give them their sim vector:
    e_vectors = {}
    for eID in range(len(episodes_info)):
        if eID not in words_by_ep:
            e_vectors[eID] = np.array(list(e_vector.values()))
    for eID, contents in words_by_ep.items():
        temp_vector = list(e_vector.values())
        for lemma, words in contents.items():
            index = list(e_vector.keys()).index(lemma)
            temp_vector[index] = sum(words.values())
        e_vectors[eID] = np.array(temp_vector)

    # similar episodes
    sim_matrix = np.zeros((len(episodes_info), len(episodes_info)))
    for idx, eID in enumerate(episodes_info):
        for idy, ieID in enumerate(list(episodes_info)[idx + 1:], idx + 1):
            if eID not in words_by_ep or ieID not in words_by_ep:
                sim_matrix[idx][idy] = 0
            else:
                sim_matrix[idx][idy] = helper.cosine_similarity(
                    e_vectors[eID], e_vectors[ieID])
    np.savetxt(edited_path + "_sim_matrix.csv", sim_matrix, delimiter=";")

    # make_images(sim_matrix, edited_path)

    sim_matrix = sim_matrix + sim_matrix.transpose()
    max_related = 5
    related = {}
    for eID in range(len(episodes_info)):
        if eID not in words_by_ep:
            continue
        related[eID] = {}
        sim = sim_matrix[eID]
        ind = np.argpartition(sim, -max_related)[-max_related:]
        ind = ind[np.argsort(sim[ind])]
        for ieID in reversed(ind):
            if ieID not in words_by_ep:
                continue
            title_temp = episodes_info[ieID]['title']
            clean_title_temp = helper.get_clean_title(title_temp, ieID, True)
            related[eID].update({clean_title_temp: sim[ieID]})

    # graphvizify.dotify()
    # graphvizify.generate_dot_file(dot, "_dot", edited_path)

    # helper.dict2json(linkpreps, "_linkpreps", edited_path)

    for eID, episode_info in episodes_info.items():
        eID = int(eID)
        # if eID <= 20:
        #     continue
        # if eID > 100:
        #     return
        title = episode_info["title"]
        clean_title = helper.get_clean_title(title, eID)
        file_path = edited_path + clean_title + ".json"
        if os.path.exists(file_path):
            with open(file_path, encoding='utf-8') as json_file:
                transcript = json.load(json_file)
        else:
            print("Missing: " + file_path)
            continue
        # convert transcript
        note = make_a_note_out_of_me(transcript, episode_info, playlist_info)
        links = {}
        if eID in words_by_ep:
            links = words_by_ep[eID]
        note += add_links(links)
        if len(links) > 0:
            note += add_related(related[eID], playlist_info['title'])
        note += add_full_text(transcript['text'], links)
        # write transcript
        clean_title = helper.get_clean_title(title, eID, True)
        filepath = obs_pl_path + "\\" + clean_title + ".md"
        with open(filepath, 'w', encoding='UTF8') as f:
            f.write(note)


def include_lexicon(edited_path, playlist_name):
    with open(edited_path + '_lexicon_similar.json', encoding='utf-8') as json_file:
        lexicon_similar = json.load(json_file)
    with open(edited_path + '_lemmacon.json', encoding='utf-8') as json_file:
        lemmacon = json.load(json_file)
    if not os.path.exists(OBSIDIAN_PATH + "\\lexicon"):
        os.makedirs(OBSIDIAN_PATH + "\\lexicon")
    for key, value in lemmacon.items():
        # handle if note already exists
        count_in_ = 0
        flexeds = []
        for flexed, count in value.items():
            count_in_ += count
            flexeds.append(flexed)
        if count_in_ < 3:
            continue
        similar = []
        if key in lexicon_similar:
            similar = lexicon_similar[key]
        note = "---\n"
        note += "count_in_" + playlist_name + ": " + str(count_in_) + "\n"

        note += "---\n"

        tags = "#lexem "

        note += "tags:: " + tags + "\n"
        comment = ""
        note += "comment:: " + comment + "\n"
        note += "\n---\n"
        note += "# Inflected words\n"
        note += "; ".join(flexeds) + "\n\n"
        note += "# similar words\n"
        note += "; ".join(similar) + "\n\n"
        # note += add_links(value)
        # write transcript
        filepath = OBSIDIAN_PATH + "\\lexicon\\" + key + ".md"

        with open(filepath, 'w', encoding='UTF8') as f:
            f.write(note)


def include_to_obsidian(input_path, playlist_name, playlist_info, episodes_info):
    obs_pl_path = OBSIDIAN_PATH + "\\" + playlist_name
    if not os.path.exists(obs_pl_path):
        os.makedirs(obs_pl_path)
    data_path, playlist_names = helper.get_data_folders()
    input_path = data_path + playlist_name + "\\"
    edited_path = input_path + "edited\\"

    playlist_info, episodes_info = helper.setup_infos(
        playlist_info, episodes_info, input_path)
    data_files, data_folders = helper.extract_file_folder(
        input_path + "edited\\")
    jsons = [filename for filename in data_files if filename.endswith(".json")]

    include_episodes(edited_path, obs_pl_path, episodes_info, playlist_info)
    # include_lexicon(edited_path, playlist_name)


def main(playlist_name='BMZ', input_path=os.getcwd(), playlist_info={}, episodes_info={}):
    include_to_obsidian(input_path, playlist_name,
                        playlist_info, episodes_info)


if __name__ == '__main__':
    main()



# Wichtig:
# Groß- und Kleinschreibung beachten
# -> handled by "case insensitive"


name = {
    # Barlow:
    r'(?<!\w)[PBVW]a{1,2}r{0,1}[ndl]{0,2}o{1,2}u{0,1}w{0,1}(?!\w)':   'Barlow',
    r'(?<!\w)[PB]a{1,2}r{0,1}[ndl]{0,2}o{1,2}u{0,1}w{0,1}r{0,1}(?!\w)':   'Barlow',
    r'(?<!\w)[vVwWmM][aA]{1,2}[nNdDlL]{1,2}[oOuU]{1,2}[mMwW]{0,1}(?!\w)':   'Barlow',
    r'(?<!\w)[vVwWbB][aAhH]{1,2}[nNdDlL]{1,2}[oOuU]{1,2}[mMwW]{0,1}(ng){0,1}(?!\w)':   'Barlow',
    r'(?<!\w)[vVwWbB][aAhH]{1,2}[nNdDlL]{1,2}[oOuU]{1,2}[mMwW]{0,1}(ng){0,1}(?!\w)':   'Barlow',
    r'\bBalot*g*(ch)*(?!(mer))': 'Barlow',
    r'\bpal[ou]\b': 'Barlow',
    r'\bBalos\b': 'Barlows',

    # Onkel Barlow:
    r'(?<!\w)unkel{0,1}(?!\w)':   'Onkel',
    'onkel Campus': 'Onkel Barlow',
    'Onkel Manu': 'Onkel Barlow',
    'onkel war lo': 'Onkel Barlow',
    'onkel war': 'Onkel Barlow',
    'UnclipAllo': "Onkel Barlow",
    r'(?<!\w)[oO]nkel{0,1} (pENNIS)*(Wadl)*(pa am)*(?!\w)':   'Onkel Barlow',
    # pENNIS? WTF Whisper? WTF Trainingssatz...
    # onkel anime # nope, das war intentional

    'Balomer,': "Barlow, mal",
    r'\bBalomarc*k*t*\W': 'Barlow mag ',

    'Uli Sebalo': "Hallo Barlow",
    r'\bH[ae]lobar*low*\b': 'Hallo Barlow',
    r'\b[nh]e[iy]*ba[dr]*low*\b': 'Hey Barlow',
    # r'\bhei{0,1}ba[dr]ow{0,1}': 'Hey Barlow',
    # r'\bhai{0,1}ba[dr]ow{0,1}': 'Hi Barlow',
    r'\b[PH][ae]ibar*low*\b': 'Hi Barlow',
    r'\bJoBalo\b': 'Jo Barlow',
    r'\bHu*bar*low*\b': 'Huhu Barlow',
    r'\bHu*[ou]nkel*bar*low*\b': 'Huhu Onkel Barlow',
    # herzlich willkommen hier ist Barlow
    'herzlich willkommen hier ist war noch': 'herzlich willkommen hier ist Barlow',
    # r'willkommen,* hier ist [^(Barlow)] mit '   :   'willkommen, hier ist Barlow mit ', # nicht mehr notwendig
}

abrev = {
    r'\[kc]atar*\b': "Cata",
    r'\bBF\.*[- ][RA]\b': "BfA",
    r'\bPvE\b': "PvE",
    r'\bPvP\b': "PvP",
    r'\bSW.Tou*r\b': "SWTOR",
    r'\btl[ ;\.]dr\b': "TLDR",
    r'\bbow\b': "WoW",
    r'\b[vw]\.*o*\.*[vw]\.*\b': "WoW",
    r'\bwod\b': "WoD",
    r'\b[mdwpb][mn]z\b': 'BMZ',
    r'\b[aA][hH][aA]*\b': "AH",
    'AH-Effekt': "Aha-Effekt",
    r'\bnfr\b': "LFR",
    r'\bDDS\b': "DDs",
    r'\beds\b': "Adds",
    r"(?<![-'])\bED\b": "Add",
    r'\b[vV][oO][tT][lL][kK]\b': "WotLK",
    'D.O.T.L.K.': 'WotLK',
    r'\b[cC]{2}S\b': 'CCs',
    r'Radar\.io': 'raider.io',
    r'Rai*d[ea]r[\. (,)]i\.*o\.*': 'raider.io',
}

words = {
    r'\bheel(?=\w)': "heal-",
    r'\bheel': "heal",
    r'\bheel': "heal",
    r'\bhelse\b': "heal",
    r'\bhefen\b': "helfen",
    r'\bguilde(?=[n\b])': "Gilde",
    r'\b(?<!Sammel)gilder\b': "Gilde",
    r'(?<=[e ])radet': "geraided",
    r'\b[pP]al+adin*\b': "Paladin",
    # leaving rate -> Rate in there, but is not always right:
    r'\brate\b': "Raid",
    r'\bgerated\b': "geraided",
    r'\b\w*Wa\w* of Draenor': "Warlords of Draenor",
    r'\b\w*Wa\w* of Draenor': "gecue",
    r'\b\w*Wa\w* of Draenor': "-ed un 'ed",
    # testing

    # testing end
    r'\bFierigkeiten\b': "Schwierigkeiten",
    'Formists of Pandaria': "vor Mists of Pandaria",

    # Cue (Hinweis) und Queue (Warteschlange)
    r"gecue.ed": "gequeued",
    r"Solo.cue": "Solo Queue",
    r"Dungeon.cue": "Dungeon Queue",
    r"cue.window": "Queue Window",

    r"Lordhawk": "Lore Talk",
    r"Lore-Talk": "Lore Talk",
    "Poiseken": 'Päuseken',

    # Latein
    "Cumhawk ergopropterhawk": 'cum hoc ergo propter hoc',
    "posthawk ergopropterhawk": 'post hoc ergo propter hoc',

    r"\bKil[' ]jae*den\b": "Kil'jaeden",
    r"\bDeb\b": "Depp",
    r'\bini\b': "Ini",
    'Mythik': "Mythic",
    r'[hH]ailer': "Heiler",
    'Edron': 'Add-on',
    'Addon': 'Add-on',
    r'\b[iI]ngenie-(?=\w)': 'ingame-',
    r'\b[iI]ngenie(?!-)\b': 'Ini',
    'Amount': 'Mount',
    r'[dD]ungen': 'Dungeon',
    'älter Scrolls': 'Elder Scrolls',
    'Cell-Run': 'Sellrun',
    r'[pP]arlor': 'Pala',
    r'\b[pP][rRoO]{0,2}t[ -][pP][aA][rR]{0,1}[lA]+[aAoP][rR]{0,1}': 'Prot Pala',
    # 'gemopper' : 'gemopper', # dafuq?
    # 'baff' : 'Buff', # nicht immer
    # heel, self hier : 'heal', 'self heal
    'Olof Goldcap': 'Null auf Goldcap',
    r'\bingame[^-]':  'ingame-'
}

acro = {
    'BMZ': {
        'Cata': ['Cataclysm'],
        'Afd': ['Alternative für Deutschland'],
        'SWTOR': ['Star Wars: The Old Republic'],
        'BMI': ['body mass index'],
        'DMB': ['Deadly Boss Mods'],
        'TBC': ['The Burning Crusade'],
        'WoW': ['World of Warcraft'],
        'LFR': ['looking for raid'],
        'LFG': ['looking for group'],
        'BG': ['battleground'],
        'RBG': ['random battleground'],
        'IO': ['Raider.IO'],
        'MC': ['Molten Core'],
        'HC': ['heroic'],
        'MVP': ['at the moment'],
        'MVP': ['most valuable player'],
        'Dev': ['Developer'],
        'aka': ['also known as'],
        'DK': ['death knights'],
        'MMO': ['massively multiplayer online'],
        'MMOG': ['massively multiplayer online game'],
        'MMORPG': ['massively multiplayer online role-playing game'],
        'DD': ['damage dealer'],
        'nHC ': ['non heroic'],
        'PvP': ['Player versus Player'],
        'PvP': ['Player versus Environment'],
        'Mob': ['mobile object'],
        # TODO: fix mistake with "Adds" -> most likely rework acro detection
        # 'Add': ['additional enemy', 'Adds', 'additional enemies'],
        'DPS': ['damage per second'],
        'CC': ['crowd control'],
        'Pot': ['potion'],
        'DoT': ['damage over time'],
        'RP': ['role-play'],
        'ERP': ['erotic role-play'],
        'RPG': ['role-playing game'],
        'HPS': ['healing per second'],
        'HoT': ['healing over time'],
        'LKW': ['Lastkraftwagen', 'LKWs', 'Lastkraftwagen'],
        'AH': ['Auktionshaus'],
        'BMAH': ['Black Market Auction House'],
        'TSM': ['TradeSkillMaster'],
        'OG': ['Orgrimmar'],
        'TLDR': ["too long; didn't read"],
        'Ini': ['Instanz'],
        'M+': ['Mythic+'],
        'BfA': ['Battle for Azeroth'],
        'WotLK': ['Wrath of the Lich King'],
        'WoD': ['Warlords of Draenor'],
        'DKP': ['dragon kill points'],
        'NFL': ['National Football League'],
        'Abo': ['Abonnement'],
        'MoP': ['Mists of Pandaria'],
        'USA': ['United States of America'],
        'DDR': ['Deutsche Demokratische Republik'],
        'GCD': ['global cooldown'],
        'AE': ['Area of Effect'],
        'AoE': ['Area of Effect'],
        'CD': ['cooldown'],
        'ARD': ['Arbeitsgemeinschaft der öffentlich-rechtlichen Rundfunkanstalten der Bundesrepublik Deutschland'],
        'ZDF': ['Zweites Deutsches Fernsehen'],
        'API': ['application programming interface'],
        'oom': ['out of mana'],
        'BRD': ['Bundesrepublik Deutschland'],
        'ToS': ['terms of service'],
        'RNG': ['random number generator'],
        'Int': ['intelligence'],
        'Def': ['defense'],
        'App': ['application'],
    }
}

for key, value in acro.items():
    acro[key] = dict(sorted(value.items(), key=lambda i: i[0].lower()))


"""
376
Steht BMZ für barlos Meinungs-Streit, Meinungs-Austauschzimmer? 
Ich sage gar nichts dazu. Also weder ein Ja 
noch ein Nein. Wenn ich jedes Mal Nein sage 
und ihr irgendwann das Richtige sagt, dann würde 
ich nicht Nein sagen und ihr wüsstet, AH, da du 
nicht Nein gesagt hast, sind wir wohl auf dem richtigen Weg. 
Eines Tages werde ich verkünden, was BMZ heißt. 
Ich muss mir noch überlegen, wann genau ich das mache. 
Vielleicht im 1000. BMZ oder so? Keine Ahnung. 
Ich habe das schon mal, glaube ich, erklärt, dass es 
für mich so ein Seriengeheimnis ist wie bei den Fernsehserien 
in den 70er, 80er, teilweise auch noch 90er Jahren,
wo es immer in jeder Serie so eine Kleinigkeit gab, 
die man irgendwie nicht wusste und die so das Geheimnis war.
"""

"""
findall(r'\bB\w* M\w* Z\w*')
Folge 488
Fünfte Frage steht BMZ für 
Barlow Mecker zurück 
oder 
Balos 
Meinungszeit. Ich weiß, dass ich darauf keine Antwort
 bekomme, aber es macht mich fuchsig. 
 Irgendwann werde ich eine Antwort geben. 
 Ich überlege gerade, wann ich die Antwort 
 gebe. Gebe ich die Antwort in einem bestimmten 
 BMZ, im 1000. zum Beispiel? Vielleicht kündige 
 ich es nicht vorher an, sodass ihr euch denkt, 
 Mensch, jetzt muss ich aber mal das 1000. BMZ 
 gucken. Oder mache ich das, wenn wir irgendeinen
 anderen Meilenstein erreichen? Zuschauerzahlen 
 oder so, oder Subs oder Abos oder so? Keine Ahnung. 
 Irgendwann werde ich es bestimmt mal verraten, 
 aber im Moment noch nicht. Vielleicht passiert es 
 auch einfach im 1142. BMZ oder so. So was, 
 was überhaupt keiner rechnet und sich dann nur 
 im Nachhinein die Leute denken, verdammt, 
 damals dieser Barlow im 488. BMZ, der ist 
 schon gesagt, im 1241. BMZ könnte er verraten, 
 wofür BMZ steht. Ich habe es gewusst und es 
 mir gebuckmarkt und es muss genau, jetzt könnt 
 ihr irgendjemand ausrechnen, welches Datum das 
 ist in dem das 1000. BMZ kommt. Irgendwann 
 verrate ich es euch. Vielleicht mal, wenn 
 ihr mal lieb seid. Und ja gut, das ist dann 
 lange nicht, bis zum 1241. BMZ. Das soll es 
 gewesen sein für heute. Macht es gut und tschüss, 
 sagt euer Onkel Barlow.

"""

"""
851

Balomarck Züge
"""

main = {
    'BMZ':   name | abrev | words
}



# controllers.py
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...extract.nlp.util_nlp import Folder
from ...core import helper


import urllib.parse
import os
import json
import re
from importlib.resources import as_file, files

LIMIT_BAG_OF_WORDS = 10
LIMIT_NAMED_ENTITIES = 5
SIM_THRESHOLD = 0.05


class ObsidianNote:
    def __init__(self, path=None, name=None, text=None, simVec=None):
        self.path = path
        self.name = name
        self.text = text or None
        self.related_note_ids = None
        if path and text is None:
            self.from_file(path)
        # if not self.iscomplete() and text:
        #     self.complete(text)
        if simVec:
            #TODO: Establish this function
            self.get_related_note_ids(simVec)

    def from_file(self, path):
        """
        Load transcript from the given path.

        Args:
            path (str): Path to the transcript file.
        """
        self.path = path
        try:
            if os.path.exists(path):
                with open(path, 'r', encoding='utf8') as f:
                    self.text = f.read()
        except Exception as e:
            print(e)

    def iscomplete(self):
        if self.text and self.path:
            return True
        return False

    def save(self, path=None):
        if path is None:
            path = self.path

        with open(path, 'w', encoding='utf8') as f:
            f.write(self.text)

    def get_name(self, filename=None, fallback=None):
        if self.name and not filename:
            return self.name
        if self.path or filename:
            name = filename or os.path.basename(self.path)
        elif fallback:
            name = fallback
        else:
            raise Exception("This Obsidian note cannot find a name")

        if os.path.splitext(name) != ".md":
            name += ".md"
        self.name = helper.get_clean_title(name, obsidian=True)
        return self.name

    def get_path(self, parent_path=None):
        if parent_path:
            self.path = os.path.join(parent_path, self.get_name())
        return self.path

    def build_note(self, text, highlights=None, links=None, name=None, related_notes=None, files=None, additional_meta_data=None, rq_scores=None):
        def dict_to_dataview(input_dict: dict[str, any], title="", join_char=", ", fromat_group="", open=True, end=True):
            res = ""
            is_empty = True
            if input_dict:
                if open:
                    res += f"%%{title}\n"
                elif title:
                    res += f"{title}\n"
                entry_counter = 0
                entry_sum = 0
                for key, value in input_dict.items():
                    if value == "" or value is None:
                        continue
                    if isinstance(value, str):
                        value = value.replace("\\", "/")
                    if fromat_group:
                        value = ', '.join(fromat_group.format(w)
                                          for w in value)
                    # TODO: turn this from a quick fix into a proper solution
                    entry_counter += 1
                    if title == "RQ Scores":
                        res += f"RQ{entry_counter}:: {value}\n"
                        entry_sum += float(value)
                    res += f"{key}:: {value}\n"
                    is_empty = False
                # TODO: turn this from a quick fix into a proper solution
                if title == "RQ Scores":
                    res += f"RQsum:: {entry_sum}\n"
                    res += f"RQave:: {entry_sum/entry_counter}\n"
                if is_empty:
                    res = ""
                elif end:
                    res += f"%%\n\n"
            return res

        title = name or self.get_name()
        # if links:
        #     for link_name, words in links.items():
        #         metadata += f"{link_name}:: {', '.join('[[{}]]'.format(w) for w in words)}\n"
        # if highlights:
        #     for highlight_name, words in highlights.items():
        #         metadata += f"{highlight_name}:: {', '.join('**{}**'.format(w) for w in words)}\n"

        # if related_notes:
        #     metadata += f"Related Notes:: {', '.join('[[{}]]'.format(n) for n in related_notes)}\n"

        # todo: append more info here
        # if additional_meta:
        # metadata += f""

        corpus = ""
        for type, path in files.items():
            path = path.replace("\\", "/")
            corpus += f"\n\n## {type}\n![[{path}]]"

        # Replace first occurrence of each word in text with a link if it's in links or highlights
        transcript = ""
        if text:
            for word in text.split():
                if word in links:
                    text = text.replace(word, f"[[{word}]]", 1)
                elif word in highlights:
                    text = text.replace(word, f"**{word}**", 1)
            transcript += f"\n\n## Transcript\n{text}"

        metadata = ""
        metadata += dict_to_dataview(links,
                                     fromat_group='[[{}]]', open=False, end=False)
        metadata += dict_to_dataview(highlights,
                                     fromat_group='**{}**', open=False, end=False)
        metadata += dict_to_dataview(related_notes,
                                     fromat_group='[[{}]]', open=False, end=False)
        if metadata:
            metadata = "\n\n%%\n" + metadata + "%%"

        fin_meta = ""
        if additional_meta_data:
            fin_meta += f"\n\n%%\n"
            for key, value in additional_meta_data.items():
                if value == "" or value is None:
                    continue
                if isinstance(value, str):
                    value = value.replace("\\", "/")
                fin_meta += f"{key}:: {value}\n"
            if fin_meta != f"%%\n":
                fin_meta += f"%%\n"
            else:
                fin_meta = ""

        fin_meta += dict_to_dataview(rq_scores, title="RQ Scores")
        rq_colon = ""
        for i in range(len(rq_scores)):
            rq_colon += f"\nRQ{i+1}:: "

        related_cols = ""
        related_sort = ""
        if rq_scores:
            pieces = []
            for i, key in enumerate(rq_scores.keys()):
                pieces.append(f"{key} as RQ{i+1}")
            rq_sum = " + ".join(rq_scores.keys())
            pieces.append(f"round({rq_sum},3) as \"Sum\"")
            related_cols = "\n" + ", ".join(pieces)
            related_sort = f"\nSORT {rq_sum} desc"

        # ORKG block:
        ORKG_data = {}
        if additional_meta_data:
            url_suffix = None
            look_at = ["doi", "title"]
            while look_at and not url_suffix:
                check = look_at.pop(0)
                if check in additional_meta_data and additional_meta_data[check]:
                    url_suffix = additional_meta_data[check].replace(
                        "/", "%2F")
            if url_suffix:
                ORKG_data[
                    "ORKG_search"] = f"https://orkg.org/search/{urllib.parse.quote(url_suffix)}"
        fin_meta += dict_to_dataview(ORKG_data, title="ORKG")

        ORKG_Document_Properties = ""
        if ORKG_data:
            for key in ORKG_data.keys():
                ORKG_Document_Properties += f"\n| {key} | `=this.{key}` |"

# Rank according to age of paper

        property_block = ""
        property_view = f"""

## Document Properties
| Property | Value |
| --- | --- |
| Date | `=this.date` |
| Authors | `=this.author` |
| DOI | `=this.doi` |
| Published in | `=this.published_in` |
| Paper URL | `=this.url` |{ORKG_Document_Properties}
	(You can update this in edit mode at the end of the document)"""

        if os.path.exists(self.path):
            with open(self.path, 'r', encoding='utf8') as f:
                text = f.read()
                pieces = text.split("\n## Contributions\n")
                if len(pieces) > 1:
                    pieces = pieces[1].split(
                        "Paper read and Contribution completed")
                if len(pieces) > 1:
                    property_block = "\n\n## Contributions\n" + pieces[0] + \
                        "Paper read and Contribution completed"

        # todo see if i can compare these property_blocks better
        if not property_block:
            property_block = f"""

## Contributions
### Contribution 1
research_problem:: 
result:: 
method:: 
material:: {rq_colon}
- [ ] Paper read and Contribution completed"""

        self.text = f"""# {title}{property_view}{property_block}{corpus}

### Related Documents:
```dataview
TABLE{related_cols}
WHERE contains(file.inlinks, this.file.link){related_sort}
```{transcript}{metadata}{fin_meta}
"""


# Directly from dictionary
#TODO: Make this package-worthy
        
# with open(os.path.join("bnw_tools","ORKG", "ResearchFields.json")) as json_file:
               
# a = files('bnw_tools.ORKG')
# b = a.joinpath('ResearchFields.json')
# c = as_file(b)

# with open(as_file((files('bnw_tools.ORKG').joinpath('ResearchFields.json')))) as json_file:

with open(files('bnw_tools.ORKG').joinpath('ResearchFields.json')) as json_file:
    meta = json.load(json_file)

RF_map = {}
for rf in list(meta.values())[0]:
    RF_map.update({rf['id']: rf['label']})
RF_map = {k: v for k, v in sorted(
    RF_map.items(), key=lambda item: int(item[0][1:]))}

# with open(os.path.join(root_stepup, "bnw_tools","ORKG", "RF_Map.json"), 'w') as outfile:
with open((files('bnw_tools.ORKG').joinpath('RF_Map.json')), 'w') as outfile:
    json.dump(RF_map, outfile, indent=4)

RF_patterns = {}
for raw_field in RF_map.values():
    pot_fields = raw_field
    if " " not in raw_field:
        continue
    fields = []
    master = ""
    patterns = []
    if " of " in raw_field:
        master = raw_field.split(" of ", 1)[0]
        pot_fields.replace(f"{master} of ", "")
        master = re.escape(master) + r"*"
    elif " and " in raw_field:
        temp = raw_field.split(" and ", 1)[1]
        if " " in temp:
            temp = " " .join(temp.split(" ", 1)[1:])
            master = re.escape(temp) + r"*"
    fields = re.split(',|/| and ', raw_field)
    fields = [field.strip(' ') for field in fields]
    if "" in fields:
        fields.remove("")
    if len(fields) > 1:
        for field in fields:
            pat = r"\b" + re.escape(field) + r"\b"
            patterns.append(pat + r".*" + master)
            patterns.append(master + r".*" + pat)
    if patterns:
        RF_patterns[raw_field] = patterns


def identify_field_weight(field, content):
    if field in content:
        return 1
    elif field in RF_patterns:
        # Is: "Electrical Engineering"
        # Field: "Electrical and Computer Engineering"
        for pattern in RF_patterns[field]:
            if re.search(pattern, content):
                return 1  # maybe weigh this less?


def update_RF(meta):
    # TODO
    # Attempt to get RF:
    weights = {
        "booktitle": 10,
        "journaltitle": 10,
        "title": 4,
        # "abstract": 1,
    }
    assume = {}
    for search_key, weight in weights.items():
        if search_key in meta:
            for RF_id, field in RF_map.items():
                # TODO:
                value = identify_field_weight(field, meta[search_key])
                if value:
                    if RF_id not in assume:
                        assume[RF_id] = 0
                    assume[RF_id] = value * weight
    assume = {k: v for k, v in sorted(
        assume.items(), key=lambda item: item[1], reverse=True)}
    highest = 0
    results = []
    for RF_id, value in assume.items():
        # TODO multiple matches?

        for result in results:
            # check if "Engineering" could be made "Computer Engineering"
            comp = identify_field_weight(RF_map[result], RF_map[RF_id])
            if comp:
                # "Engineering" is in "Computer Engineering"
                results.remove(result)
                results.append(RF_id)
                highest = value
                continue

        if value >= highest:
            if RF_id not in results:
                results.append(RF_id)
            highest = value

            # 'R137': 10,     "R137": "Numerical Analysis/Scientific Computing",
            # 'R194': 10,     "R194": "Engineering",
            # 'R237': 10,     "R237": "Electrical and Computer Engineering",
            # 'R241': 10,     "R241": "Electrical and Electronics",
            # 'R254': 10,     "R254": "Materials Science and Engineering",
            # 'R201': 4     "R201": "Structures and Materials",
    # If you want to be able to read the results
    # results = [RF_map[x] for x in results]
    if results:
        # data.update({f"paper:{key}": "orkg:" + "; orkg:".join(results)})
        # data.update({f"paper:{key}": "; ".join(results)})
        # data.update({f"paper:{key}" : f"orkg:{results[-1]}"})
        meta.update({f"research_field": f"{results[-1]}"})
        meta.update({f"research_field_label": f"{RF_map[results[-1]]}"})
    return meta


def folder(folder: Folder, limit_ns=LIMIT_BAG_OF_WORDS, limit_ne=LIMIT_NAMED_ENTITIES, force=False):
    # loop over each file in the directory
    # build Entry Node
    # Sort all files into subfolders

    for idr, mr in enumerate(folder.media_resources):
        if not mr.obsidian_note:
            mr.add_obsidian_note()
        mr.obsidian_note.get_name(mr.original_name, fallback=mr.basename)
        mr.obsidian_note.get_path(folder.notes.path)

        # Create the Obsidian directory if it does not exist,

        bag_of_words = list(mr.nlp_analysis.bag_of_words.tf_idf.keys())[
            :LIMIT_BAG_OF_WORDS + LIMIT_NAMED_ENTITIES]
        named_entities = list(mr.nlp_analysis.named_entities.get().keys())[
            :LIMIT_NAMED_ENTITIES]

        for name in named_entities:
            lookfor = set([name.lower()]).union(name.lower().split())
            for word in lookfor:
                if word in bag_of_words:
                    bag_of_words.remove(word)
        bag_of_words = bag_of_words[:LIMIT_BAG_OF_WORDS]

        links = {}
        highlights = {}

        if named_entities:
            links["named_entities"] = named_entities

        if bag_of_words:
            highlights["bag_of_words"] = bag_of_words

        files = {}
        if mr.audio_file:
            files["Audio"] = os.path.basename(mr.audio_file.path)
        if mr.pdf:
            files["PDF"] = os.path.relpath(mr.pdf.path, folder.path)

        simVec = folder.sim_matrix[idr]
        top_indices = simVec.argsort()[-6:-1]
        related_notes = []
        for i in top_indices:
            # check if similarity is high enough
            if simVec[i] < SIM_THRESHOLD:
                break
            cr = folder.media_resources[i]
            related_notes.append(cr.obsidian_note.get_name(cr.original_name))
        related_notes = {"Related Notes": related_notes}
        text = None
        if mr.transcript:
            text = mr.transcript.text

        meta = {}
        if mr.pdf and mr.pdf.metadata:
            meta.update(mr.pdf.metadata)
        if mr.info:
            block = mr.info.__dict__
            del block['thumbnail_urls']
            del block['cap_codes']
            meta.update(block)

        if meta:
            # research field:
            # update_RF(meta)

            # date
            if "date" in meta:
                values = meta["date"].split("-")
                meta.update({"publication_year": values[0]})
                if len(values) > 1:
                    meta.update({"publication_month": values[1]})

            # date
            for key in ["booktitle", "journaltitle"]:
                if key in meta:
                    meta.update({"published_in": meta[key]})
                    break
        rq_scores = {}
        try:
            if folder.rq and folder.rq.research_questions:
                for id_rq, rq in enumerate(folder.rq.research_questions):
                    rq_scores.update(
                        {rq.tag: round(folder.rq_sim_mat[idr][id_rq], 3)})
        except Exception as e:
            print(e)

        name = mr.original_name

        if mr.pdf and mr.pdf.metadata:
            if 'title' in mr.pdf.metadata:
                name = mr.pdf.metadata['title']
        elif mr.info:
            if hasattr(mr.info, 'title'):
                name = mr.info.title

        mr.obsidian_note.build_note(
            text, links=links, highlights=highlights, name=name, related_notes=related_notes, files=files, additional_meta_data=meta, rq_scores=rq_scores)

        # create the markdown file
        mr.obsidian_note.save()



import json
import datetime

# read the json file
with open("results/large_v2.json", "r") as f:
    data = json.load(f)

for i in range(len(data["segments"])):
    # get the start time
    start = data["segments"][i]["start"]
    # start time from ms to hh:mm:ss,xxx
    start = datetime.datetime.fromtimestamp(start / 1000.0).strftime('%H:%M:%S,%f')
    # get the end time
    end = data["segments"][i]["end"]
    # end time from ms to hh:mm:ss,xxx
    end = datetime.datetime.fromtimestamp(end / 1000.0).strftime('%H:%M:%S,%f')
    # get the text
    text = data["segments"][i]["text"][1:]
    bytes = text.encode('latin-1')
    text = bytes.decode('raw_unicode_escape')
    # write the vtt file
    with open("results/large-v2.vtt", "a") as f:
        f.write(str(i) + "\n")
        f.write(str(start) + " --> " + str(end) + "\n")
        f.write(str(text) + "\n\n")


from core import helper
import os
import numpy

FLUSH_OUT_RELICS = True


# ---- Globals ---- #


# --- 1. general --- #
# Replacements to be done from HTML to readable text

replace_dict = {
    "&nbsp;":   "",
    "&lt;":   "<",
    "&gt;":   ">",
    "&amp;":   "&",
    "&euro;":   "€",
    "&pound;":   "£",
    "&quot;":   "“",
    "&apos;":   "‘",
    "\\u00FC":   "ü",
    # TODO: further äö and others should be added
    # https://www.cl.cam.ac.uk/~mgk25/ucs/quotes.html
    "”":    '"',
    "“":    '"',
    "&#39;": "'",
}


# --- 2. specific --- #
# order keys should be read in (for sort_dict function)
key_order = (
    'id',
    'date',
    'runtime',
    'name',
    'title',
    'description',
    'author_type',
    'author_name',
    'author_url',
    'publisher',
    'language',
    '@type',
    'accessMode',
    'url',
    'image'
)

dict_playlist_YT = {
    "make_boxes":   '</yt-formatted-string></h3><div id="publisher-container" class="style-scope ytd-playlist-panel-renderer"><yt-formatted-string class="byline-title style-scope ytd-playlist-panel-renderer complex-string" ellipsis-truncate="" hidden="" ellipsis-truncate-styling="" title="',
    "box_begin":   'omplex-string" ellipsis-truncate="" hidden="" ellipsis-truncate-styling="" title="',
    "box_end":   '</a></yt-formatted-string><div class="index-message-wrapper style-scope ytd-playlist-panel-renderer"><span class="index-message style-scope ytd-playlist-panel-renderer" hidden="">',
    "title":   '" has-link-only_=""><a class="yt-simple-endpoint style-scope yt-formatted-string" spellcheck="false" href="',
    "url":   '" dir="auto">',
    "void":   '</a></yt-formatted-string><ytd-badge-supported-renderer class="style-scope ytd-playlist-panel-renderer" disable-upgrade="" hidden=""></ytd-badge-supported-renderer><yt-formatted-string class="publisher style-scope ytd-playlist-panel-renderer complex-string" ellipsis-truncate="" link-inherit-color="" ellipsis-truncate-styling="" title="',
    "author_name":   '" has-link-only_=""><a class="yt-simple-endpoint style-scope yt-formatted-string" spellcheck="false" href="',
    "author_url":   '" dir="auto">',
}

dict_episode_YT = {
    "make_boxes": '<ytd-video-meta-block class="playlist',
    "box_begin":   '<a id="video-title" class="yt-simple-endpoint style-scope ytd-playlist-video-renderer" href=',
    "box_end":   '">\n',
    "title":   ' title=',
    "url":   ''
}


# Dict with episode keys to be found and values as extraction markers (for cut_out function)
dict_episode = {
    "final_destination_void": "</span></p></div></div>",
    "runtime":   '</p><p class="Type__TypeElement-sc-goli3j-0 hGXzYa _q93agegdE655O5zPz6l"><span class="UyzJidwrGk3awngSGIwv">',
    "date":   '</div><div class="qfYkuLpETFW3axnfMntO"><p class="Type__TypeElement-sc-goli3j-0 hGXzYa _q93agegdE655O5zPz6l">',
    "void":   '</p></div><div',
    "description": "</div></a></div></div><div class=\"upo8sAflD1byxWObSkgn\"><p class=\"Type__TypeElement-sc-goli3j-0 hGXzYa LbePDApGej12_NyRphHu\">",
    "title":   '><div class="Type__TypeElement-sc-goli3j-0 kUtbWF bG5fSAAS6rRL8xxU5iyG" data-testid="episodeTitle\">',
    "url":   'href='
}

# Dict with playlist keys to be found and values as extraction markers (for cut_out function)
dict_playlist = {
    "final_destination_void": "}</script><link rel=",
    "language":   ',"inLanguage":',
    "accessMode":   ',"accessMode":',
    "image":   ',"image":',
    "author":   ',"author":',
    "publisher":   ',"publisher":',
    "description":   ',"description":',
    "title":   ',"name":',
    "url":   ',"url":',
    "@type":   ',"@type":'
}


# ---- Functions ---- #


def sort_dict(dict, order_list):
    temp = {}
    for k in order_list:
        if k in dict:
            temp[k] = dict[k]
    return temp


def sort_episodes(episodes):
    first = int(episodes[0]['date'].split('-', 1)[0])
    # reminder: we cannot fine-sort every entry, since the date is not stored precisely
    lid = len(episodes)-1
    second = int(episodes[lid]['date'].split('-', 1)[0])
    if second > first:
        return episodes
    else:
        temp = {}
        for idx, k in enumerate(reversed(list(episodes.keys()))):
            temp[idx] = episodes[k]
        return temp


def Spotify2dict(index, input_path, filename, playlist_name):
    # Cut in between episode and playlist info
    split_index = index.split("alt=")

    # extract playlist info
    playlist_info = helper.cut_out(split_index[0], dict_playlist)

    # remove unneeded rows
    split_index.pop(0)
    split_index.pop(0)

    # cutting out "neue Folge"
    episodes = {}
    for idx, episode in enumerate(split_index):
        sub = episode.split("Neue Folge\"></span>")
        if len(sub) > 1:
            episode = sub[1]
        episode_info = helper.cut_out(episode, dict_episode)
        if 'description' in episode_info:
            episode_info["description"] = episode_info["description"].split("Werbung:", 1)[
                0]
        episodes[idx] = episode_info
    # m = re.search('(?<=episodeTitle">).+?(?=<\/div>)', index)
    # m.group(0)

    # TODO: Ideen
    # replace_dict benötigt ausnahme für Links
    # Werbung raus
    # Link
    # Nummer
    # add language and other data to episodes (can we do this? from podcast alone? redundand? what about changes?)
    # ID
    # Convert to Wiki page
    # Sortable
    # Kategorie
    # Struktur
    # URL von neue Folgeklappt noch nicht

    # ---- Clean ---- #
    # sort_dict(playlist_info, key_order)
    playlist_info = helper.clean_dict(playlist_info, playlist_name)
    for key, value in episodes.items():
        episodes[key] = helper.clean_dict(value, playlist_name)

    # ---- Sort ---- #
    # sort_dict(playlist_info, key_order)
    playlist_info = sort_dict(playlist_info, key_order)
    for key, value in episodes.items():
        episodes[key] = sort_dict(value, key_order)
    episodes = sort_episodes(episodes)
    return playlist_info, episodes


def get_playlist_info_YT(var_str):
    temp = {}
    for key, value in dict_playlist_YT.items():
        if key == "make_boxes" or key == "box_begin":
            var_str = var_str.split(value, 1)[1]
        elif key == "box_end":
            var_str = var_str.split(value, 1)[0]
        else:
            group = var_str.split(value, 1)
            if 'void' not in group[0] and len(group[0]) > 0:
                temp[key] = group[0]
            var_str = group[1]
    return temp

# convert from youtube index to dictionary


def YouTube2dict(index, input_path, filename, playlist_name):
    boxes = index.split(dict_episode_YT['make_boxes'])
    episode_info_YT = {}
    playlist_info_YT = {}
    # Todo: Extract youtube Link to playlist
    for idx, box in enumerate(boxes):
        if idx == 0:
            playlist_info_YT = get_playlist_info_YT(box)
        box = box.split(dict_episode_YT['box_begin'])[-1]
        box = box.split(dict_episode_YT['box_end'])[0]
        episode = {}
        rest = box.split(dict_episode_YT['title'])
        if len(rest) > 1:
            episode['title'] = helper.remove_quot(rest[1])
            episode['url'] = helper.remove_quot(rest[0])
            episode_info_YT[idx] = episode
    return playlist_info_YT, episode_info_YT


# convert from dictionary to json


def search_for_title(try_this, episodes_info, playlist_name):
    for idx, eID in enumerate(episodes_info):
        comp = helper.title_mine(episodes_info[eID]['title'], playlist_name)
        is_match = helper.compare_titles(try_this, comp, playlist_name)
        if is_match >= 1:
            return idx
        # if 'try_pl_name' in try_this and 'try_pl_name' in comp:   Playlists should always be the same
    # Currently limited to only compare longer titles, otherwise may be unprecise
    #! Very time expensive
    # if len(title.lower()) > 10:
    #     for idx, eID in enumerate(episodes_info):
    #         comp = helper.title_mine(episodes_info[eID]['title'])
    #         sim = helper.similar(title, episodes_info[eID]['title'])
    #         if sim:
    #             return idx
    return -1

# handle the difference between all the different input sources
# Todo: for example: add missing episodes, missing links, titles, add different links per source, ...


def handle_diff_ep(list_var, playlist_name):
    for element in list_var:
        if not element:
            list_var.remove(element)
    longest = 0
    for idx, element in enumerate(list_var):
        if len(element) > len(list_var[longest]):
            longest = idx
    specials = 0
    # Step 1:
    # run through this once and allign every possible title
    # possible results:
    # fID >= 0
    #       wir haben ein Match gefunden und es ist an der Stelle fID
    # nachher prüfen, ob alle fIDs verwendet wurden ->
    #       sonst haben die kürzeren eine folge, die die längere nicht hat
    #       Dann die nachbarn checken und herausfinden, wo diese Folge in longest eingefügt werden muss
    # fID == -1
    #       wir haben kein Match gefunden
    #           diese folge existiert nur in longest

    # Phase 1: alle plID ausprobieren
    #       Ergebnis:
    # input: playlist_ID, eID, specials
    # output: fID

    # Phase 2: alle playlist_ID_shorter prüfen, die keine fID haben
    # input: fID, shorter_playlist
    # output: missing_pl_id = [prior_neighbor, later_neighbor]

    #
    # Wenn eID Sprung macht, aber kein Special dabei war: ring a bell
    #
    order = [0] * len(list_var[longest])
    matched = []
    for episode_infos in list_var:
        matched_counters = []
        for each in episode_infos:
            matched_counters.append([])
        matched.append(matched_counters)
    # print('order:')
    last_eID = 0
    for plID, episode in list_var[longest].items():
        # TODO: find out why suddenly read as string -> saved in dict -> find out why
        plID = int(plID)
        title = episode['title']
        try_this = helper.title_mine(title, playlist_name)
        try_this['plID'] = str(plID)
        # try_this['plID'] = plID
        # if the ID is missing from the main Episode
        if 'eID' in try_this:
            try:
                last_eID = int(try_this['eID'])
            except:
                last_eID += 1
        elif 'special' not in try_this:
            last_eID += 1
            try_this['eID'] = str(last_eID)
            # try_this['eID'] = str(plID+1)
            # try_this['eID'] = plID+1
        findings = [plID]
        for listID, element in enumerate(list_var):
            if listID != longest:
                # Make new dicts and resort them here
                fID = search_for_title(
                    try_this, list_var[listID], playlist_name)
                if fID >= 0:
                    matched[listID][fID].append(plID)
                findings.append(fID)
        order[plID] = findings
        # print('\t' + ',\t'.join(str(e) for e in findings), flush=True)
        title = episode['title']

    # handle Miss-Matches

    multi_match = []
    # Case 1: Multi-matches
    print('Case 1: Multi-matches:')
    for idx, element in enumerate(list_var):
        if idx != longest:
            for fID, episode in enumerate(element):
                if len(matched[idx][fID]) > 1:
                    entry = {
                        'listID':   idx,
                        'eID':   fID,
                        'matches':   matched[idx][fID]
                    }
                    multi_match.append(entry)
                    print(list_var[idx][fID]['title'])
                    print('[' + str(idx) + '][' + str(fID) + ']\t' + str(len(matched[idx][fID])) +
                          ' matches:\t' + ',\t'.join(str(e) for e in matched[idx][fID]), flush=True)
                    for id in matched[idx][fID]:
                        print('\t\t\t\t' + list_var[longest][id]['title'])

    zero_match = []
    # Case 2: 0-Matches
    print('Case 2: 0-Matches:')
    for idx, element in enumerate(list_var):
        if idx != longest:
            for fID, episode in enumerate(element):
                if not matched[idx][fID]:
                    entry = {
                        'title':   list_var[idx][fID]['title'],
                        'listID':   idx,
                        'eID':   fID,
                        'matches':   matched[idx][fID],
                    }
                    zero_match.append(entry)
                    print(entry['title'])
                    print('[' + str(idx) + '][' + str(fID) + ']\t' + str(len(matched[idx][fID])) +
                          ' matches:\t' + ',\t'.join(str(e) for e in matched[idx][fID]), flush=True)
    # TODO: Remove short term fix

    # Case 1: Duplicate Upload:
    # Status: Working!
    deletes = []
    for idx, missing_match in enumerate(zero_match):
        # look if repetition:
        eID = missing_match['eID']
        listID = missing_match['listID']
        title = missing_match['title']
        try_this = helper.title_mine(title, playlist_name)
        clone_score = helper.compare_titles(try_this, try_this, playlist_name)
        for idy in range(1, 3+1):
            try:
                comp = helper.title_mine(
                    list_var[listID][eID-idy]['title'], playlist_name)
                score = helper.compare_titles(try_this, comp, playlist_name)
                if score == clone_score:
                    print('[' + str(listID) + '][' + str(eID) + ']' +
                          ' is a copy of [' + str(listID) + '][' + str(eID-idy) + ']', flush=True)
                    deletes.append(idx)
                    break
            except:
                pass
    for x in reversed(deletes):
        del zero_match[x]

    # Case 2: Ment for zero_match:
    # Status: Working!
    deletes = []
    for mmID, entry in enumerate(multi_match):
        delete_here = []
        # entry eID from list listID has to many matches from list longest, namely matches
        listID = entry['listID']
        eID = entry['eID']
        ep_prev = list_var[listID][eID-1]['title']
        episode = list_var[listID][eID]['title']
        ep_next = list_var[listID][eID+1]['title']
        episode_triple = [ep_prev, episode, ep_next]
        matched_ep_titles = []
        for match in entry['matches']:
            matched_ep_prev = list_var[longest][match-1]['title']
            matched_episode = list_var[longest][match]['title']
            matched_ep_next = list_var[longest][match+1]['title']
            matched_ep_titles.append(
                [matched_ep_prev, matched_episode, matched_ep_next])
        # Case 1:
        match_score_list = []
        for matched_triple in matched_ep_titles:
            score = 0
            for x in range(0, 3):
                try_this = helper.title_mine(episode_triple[x], playlist_name)
                comp = helper.title_mine(matched_triple[x], playlist_name)
                score += helper.compare_titles(try_this, comp, playlist_name)
            match_score_list.append(score)
        hightest = 0
        for score in match_score_list:
            if score > hightest:
                hightest = score
        for match_ID, score in enumerate(match_score_list):
            if score != hightest:
                main_eID = entry['matches'][match_ID]
                for idy, missing_match in enumerate(zero_match):
                    if missing_match['title'] != matched_ep_titles[match_ID][1]:
                        continue
                    if zero_match[idy]['listID'] != listID:
                        continue
                    pot_eID = zero_match[idy]['eID']
                    pot_ep_prev = list_var[listID][pot_eID-1]['title']
                    pot_episode = list_var[listID][pot_eID]['title']
                    pot_ep_next = list_var[listID][pot_eID+1]['title']
                    pot_title = [pot_ep_prev, pot_episode, pot_ep_next]
                    score = 0
                    for x in range(0, 3):
                        try_this = helper.title_mine(
                            pot_title[x], playlist_name)
                        comp = helper.title_mine(
                            matched_ep_titles[match_ID][x], playlist_name)
                        score += helper.compare_titles(try_this,
                                                       comp, playlist_name)
                    if score >= 2:
                        # update order (!most important)
                        order[main_eID][listID] = pot_eID
                        # update matched (optional, for better overview)
                        matched[listID][pot_eID].append(main_eID)
                        del zero_match[idy]
                        # mark for update here
                        delete_here.append(match_ID)
                        print('[' + str(listID) + '][' + str(pot_eID) + ']' + ' is matched to [' + str(
                            main_eID) + '] instead of [' + str(listID) + '][' + str(eID) + ']', flush=True)
                        break
        for x in reversed(delete_here):
            del entry['matches'][x]
        if len(entry['matches']) == 1:
            deletes.append(mmID)
    for x in reversed(deletes):
        del multi_match[x]

    # Case 3: Completely missplaced duplicates
    deletes = []
    for idx, missing_match in enumerate(zero_match):
        # look if repetition:
        eID = missing_match['eID']
        listID = missing_match['listID']
        title = missing_match['title']
        try_this = helper.title_mine(title, playlist_name)
        clone_score = helper.compare_titles(try_this, try_this, playlist_name)
        for exID in list_var[listID]:
            if exID != eID:
                episode = list_var[listID][exID]
                comp = helper.title_mine(episode['title'], playlist_name)
                score = helper.compare_titles(try_this, comp, playlist_name)
                if score == clone_score:
                    print('[' + str(listID) + '][' + str(eID) + ']' +
                          ' is a copy of [' + str(listID) + '][' + str(exID) + ']', flush=True)
                    # This is way to time consumative, but at this point: just tell the spotify / whatever person to delete the duplicate.
                    # lookaround: which fits better:
                    # Done: Option to delete the existing one and place this one in its place:
                    # eg 0 1 2 3 4 50 5 6 7 8 .. 48 49 50 51
                    # currently first 50 would be kept
                    # fixed

                    # Check around existing
                    start_ID = exID
                    error_existing = 0
                    pre_c = min([3, min([exID, int(eID)])])
                    post_c = min([3, len(list_var[listID]) +
                                 1 - max([exID, int(eID)])])
                    for pre in range(1, pre_c+1):
                        check2 = helper.title_mine(
                            list_var[listID][start_ID-pre]['title'], playlist_name)
                        error_existing += (int(comp['eID']) +
                                           pre) - int(check2['eID'])
                    for post in range(1, post_c+1):
                        check2 = helper.title_mine(
                            list_var[listID][start_ID+post]['title'], playlist_name)
                        error_existing += int(check2['eID']) - \
                            (int(comp['eID']) + post)

                    # Check around missing
                    start_ID = int(eID)
                    error_missing = 0
                    # previous: 3
                    for pre in range(1, pre_c+1):
                        check2 = helper.title_mine(
                            list_var[listID][start_ID-pre]['title'], playlist_name)
                        error_existing += (int(try_this['eID']) +
                                           pre) - int(check2['eID'])
                    # following: 3
                    for post in range(1, post_c+1):
                        check2 = helper.title_mine(
                            list_var[listID][start_ID+post]['title'], playlist_name)
                        error_existing += int(check2['eID']) - \
                            (int(try_this['eID']) + post)
                    if error_existing > error_missing:
                        # existing was the faulty duplicate
                        # matched[]
                        print('[' + str(listID) + '][' + str(exID) + ']' +
                              'may be seriously missplaced. It\'s copy is better suited and will be kept.')
                        existing_matches = matched[listID][exID]
                        if len(existing_matches) != 1:
                            print(
                                'unresolvable error at multi_match  [' + str(listID) + '][' + str(exID) + ']')
                        order[existing_matches[0]][listID] = eID
                        matched[listID][eID] = existing_matches
                        matched[listID][exID] = []
                    deletes.append(idx)

                    break
    for x in reversed(deletes):
        del zero_match[x]

    # Case 4: Newcomer -> die wollen wir behalten
    deletes = []
    inserted = 0
    for idx, missing_match in enumerate(zero_match):
        # Check if they had the missing one
        eID = int(missing_match['eID'])
        listID = missing_match['listID']
        title = missing_match['title']
        try_this = helper.title_mine(title, playlist_name)

        # previous episode
        prev_match = int(matched[listID][eID-1][0])
        prev_ep = list_var[listID][eID-1]
        prev_ep_title = prev_ep['title']
        prev_split = helper.title_mine(prev_ep_title, playlist_name)
        if 'eID' in prev_split:
            prev_mined_ID = int(prev_split['eID'])

        # next episode
        next_match = int(matched[listID][eID+1][0])
        next_ep = list_var[listID][eID+1]
        next_ep_title = next_ep['title']
        next_split = helper.title_mine(next_ep_title, playlist_name)
        if 'eID' in next_split:
            next_mined_ID = int(next_split['eID'])

        this_is_it = False
        # Todo: handle exception if no ['eID'] in prev_split or next_split
        if prev_mined_ID + 1 < next_mined_ID:
            this_is_it = True
        elif int(eID) == prev_mined_ID + 1:
            this_is_it = True
        elif int(eID) == next_mined_ID - 1:
            this_is_it = True

        if this_is_it:
            position = int(prev_match) + 1 + inserted

        # See if thing has matches
        findings = [-1]
        cur_LID = listID
        for listID, element in enumerate(list_var):
            if listID != longest or listID == cur_LID:
                fID = search_for_title(
                    try_this, list_var[listID], playlist_name)
                if fID >= 0:
                    matched[listID][fID].append(plID)
                findings.append(fID)
        order.insert(position, findings)
        deletes.append(idx)
        inserted += 1
        print('[' + str(cur_LID) + '][' + str(eID) +
              '] has been inserted at order[' + str(position) + '].')

    for x in reversed(deletes):
        del zero_match[x]

    helper.dict2json(order, "order")
    helper.dict2json(matched, "matched")
    for idx, episode_infos in enumerate(list_var):
        helper.dict2json(list_var[idx], "episode_infos_" + str(idx))
    helper.dict2json(zero_match, "zero_match")
    helper.dict2json(multi_match, "multi_match")

    # for row in order:
    #     print('\t' + ',\t'.join(str(e) for e in row))

    # TODO: Sort according to order
    episode_info = {}
    for main_ID, element in enumerate(order):
        temp_episode = {}
        episode_info_parts = []
        for listID, id_in_list in enumerate(element):
            if id_in_list >= 0:
                episode_info_parts.append(list_var[listID][id_in_list])
        episode = handle_diff(episode_info_parts)
        episode_info[main_ID] = episode
    return episode_info


def handle_diff(pieces):
    unified = {}
    used = []
    for key in key_order:
        candidates = []
        for main_ID, element in enumerate(pieces):
            if key in element:
                candidates.append(element[key])
        if len(candidates) > 0:
            used.append(key)
            if len(candidates) == 1:
                unified[key] = candidates[0]

            elif key == 'url':
                # TODO: Handle Spotify+Youtube URL
                Spotify = False
                YouTube = False
                # flush out multi-entries
                temp = []
                for candidate in candidates:
                    temp_list = candidate.split(' ; ')
                    for e in temp_list:
                        temp.append(e)
                candidates = temp
                winners = []
                for candidate in candidates:
                    if "spotify.com" in candidate:
                        if not Spotify:
                            Spotify = True
                            winners.append(candidate)
                    elif "youtube.com" in candidate:
                        if not YouTube:
                            YouTube = True
                            winners.append(candidate)
                    else:
                        winners.append(candidate)
                winner = ' ; '.join(winners)
                unified[key] = winner
            else:
                winner = ""
                print(
                    str(main_ID) + ':\t' + '\tor\t'.join(str(e) for e in candidates), flush=True)
                # just pick the longest
                for candidate in candidates:
                    if len(candidate) > len(winner):
                        winner = candidate
                unified[key] = winner

    for main_ID, element in enumerate(pieces):
        unused = []
        for key in element:
            if key not in used:
                unused.append(key)
        if len(unused) > 0:
            print('unused: ' ', '.join(unused), flush=True)
    return unified


def extract_html_info(input_path, playlist_name, playlist_info, episodes_info):
    data_files, data_folders = helper.extract_file_folder(input_path)
    data_files = helper.show_newest_files(input_path, data_files)

    for file in data_files:
        if ".html" in file:
            break
        if "episodes_info.json" == file:
            # We are up to date
            return

    # --- 1. Setup --- #
    # Log file

    # check what info is available

    # Read existing info

    # If major changes are made:
    if not FLUSH_OUT_RELICS:
        playlist_info, episodes_info = helper.setup_infos(
            playlist_info, episodes_info, input_path)

    for filename in data_files:
        if filename.split('.')[-1] == 'html':
            HTMLFile = open(input_path + filename, "r", encoding="utf8")
            index = HTMLFile.read()
            # convert from HTML to readable text
            # TODO: Make exceptions for elements within links !
            for key in replace_dict:
                index = index.replace(key, replace_dict[key])
            # handle different sources
            if filename == "YouTube.html":
                playlist_info_YT, episodes_info_YT = YouTube2dict(
                    index, input_path, filename, playlist_name)
            elif filename == "Spotify.html":
                playlist_info_SF, episodes_info_SF = Spotify2dict(
                    index, input_path, filename, playlist_name)

    tempList = [playlist_info]
    if 'YouTube.html' in data_files:
        helper.dict2json(playlist_info_YT, "playlist_info_YT", input_path)
        tempList.append(playlist_info_YT)
    if 'Spotify.html' in data_files:
        helper.dict2json(playlist_info_SF, "playlist_info_SF", input_path)
        tempList.append(playlist_info_SF)
    if len(tempList) > 1:
        playlist_info = handle_diff(tempList)
    else:
        playlist_info = tempList[0]

    tempList = [episodes_info]
    if 'YouTube.html' in data_files:
        helper.dict2json(episodes_info_YT, "episodes_info_YT", input_path)
        tempList.append(episodes_info_YT)
    if 'Spotify.html' in data_files:
        helper.dict2json(episodes_info_SF, "episodes_info_SF", input_path)
        tempList.append(episodes_info_SF)

    if len(tempList) > 1:
        episodes_info = handle_diff_ep(tempList, playlist_name)
    else:
        episodes_info = tempList[0]
    del tempList

    # write files
    helper.dict2json(playlist_info, "playlist_info", input_path)
    helper.dict2json(episodes_info, "episodes_info", input_path)

    return (playlist_info, episodes_info)


def main(data_pl_path, pl_n, playlist_info, episodes_info):
    print('extract_html_info', flush=True)
    extract_html_info(data_pl_path, pl_n, playlist_info, episodes_info)

    # csv2table(input_path, lookupCSV(input_path))


if __name__ == '__main__':
    main()



import yaml
import os
import sys

# todo: make this import cleaner
sys.path.append(os.getcwd())

import core.helper as helper
import publish.graphvizify as graphvizify

MAX_LINKS = 5
MIN_IMPORTANCE = 5
MAX_CHAR_LEN_NAME = 100

tg = {}
helper.json2dict(tg, "transitions_grouped", "data\ignore")


def super_clean_title(title):
    clean_title = title.replace("/","[SLASH]")
    clean_title = helper.get_clean_title(clean_title, obsidian=True)
    clean_title = clean_title.replace("[SLASH]","/")
    if clean_title.startswith("https"):
        clean_title = clean_title[5:]
    while clean_title.startswith("/"):
        clean_title = clean_title[1:]
    if clean_title.startswith("www."):
        clean_title = clean_title[4:]
    while clean_title.endswith("/"):
        clean_title = clean_title[:-1]
    return clean_title

nodes = []
edges = []

node_ID_dict = {}
nID_counter = 0

def get_nID(text):
    global nID_counter
    if text in node_ID_dict:
        return node_ID_dict[text]
    else:
        node_ID_dict[text] = nID_counter
        nID_counter += 1
        return node_ID_dict[text]

for key, value in tg.items():
    clean_title = super_clean_title(key)
    if len(clean_title) > MAX_CHAR_LEN_NAME:
        continue
    Obs_path = "H:/Meine Ablage/Viz_Obsidian/actions"
    note = ""
    counter = 0
    sum = value['sum']
    note += "sum"
    for action, weight in value['actions'].items():
        if len(action) > MAX_CHAR_LEN_NAME:
            continue
        traffic = round(sum*weight)
        if action != key and traffic > MIN_IMPORTANCE:
            target = super_clean_title(action)
            if target not in node_ID_dict:
                nodes.append([get_nID(target), target])
            edges.append([get_nID(clean_title), get_nID(target), traffic])
            note += f"[[{target}]]:: {weight}\n"

            counter += 1
            # if counter == MAX_LINKS:
            #     break
    if counter == 0:
        continue
    else:
        nodes.append([get_nID(clean_title), clean_title])
    filepath = Obs_path + "/" + clean_title + ".md"
    pieces = clean_title.split("/")
    if len(pieces)>1:

        os.makedirs(filepath.rsplit("/",1)[0], exist_ok=True)
        a = 1
    try:
        with open(filepath, 'w', encoding='UTF8') as f:
            f.write(note)
    except Exception as e:
        print(e)

graphvizify.generate_dot_file(nodes, edges, "actions", Obs_path.replace("/actions", "/graph"), image_format = "png")


import spellcheck.barlow as barlow

main_dict = {
    'regular':   {
        r'Nr\.*(?= \d)': 'Nummer',
        r'(?<=\d)%ig': '-prozentig',
        r'(?<=\d)%': ' Prozent',
        # '......' : ' ', # Gedankenpause
        # '...' : ' ', # auslaufende Gedanken
        r'\bmanchemal\b': 'manchmal',
        r'곳' : '',
        r'수' : '',
        r'가' : '',
        r'물' : '',
        r'т' : '',
        r'а' : '',
    }
} | barlow.main

acro = barlow.acro


# Work in Progress
from __future__ import annotations
from typing import TYPE_CHECKING

import json
import wordcloud
from wordcloud import WordCloud, ImageColorGenerator
from wordcloud import WordCloud, STOPWORDS
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import os
import cv2


if TYPE_CHECKING:
    from ..extract.nlp.util_nlp import Folder


def generate_wordcloud(dict_word_count, output_file_path='wordcloud.png', width=1920, height=1080):
    """
    Generates a wordcloud as a PNG image from a dictionary containing word count information.

    Args:
        dict_word_count: A dictionary containing word count information.
        output_file_path: The path and file name to save the generated image. Default is 'wordcloud.png'.
        width: The width of the generated image. Default is 1920.
        height: The height of the generated image. Default is 1080.

    Returns:
        None.
    """
    if not dict_word_count:
        return None

    if not output_file_path.endswith(".png"):
        output_file_path += ".png"
    wc = WordCloud(width=width, height=height)
    wc.generate_from_frequencies(dict_word_count)
    wc.to_file(output_file_path)

    new = min(height, width)
    wc2 = WordCloud(width=new, height=new)
    wc2.generate_from_frequencies(dict_word_count)
    wc2.to_file(output_file_path.replace(".png", "_square.png"))


def generate_wordcloud_mask(dict_word_count, mask_path, output_file_path='wordcloud_mask.png', width=1920, height=1080):
    """
    Generates a wordcloud as a PNG image from a dictionary containing word count information and a mask image.

    Args:
        dict_word_count: A dictionary containing word count information.
        mask_path: The path of the mask image to be used.
        output_file_path: The path and file name to save the generated image. Default is 'wordcloud_mask.png'.
        width: The width of the generated image. Default is 1920.
        height: The height of the generated image. Default is 1080.

    Returns:
        None.
    """
    if not output_file_path.endswith(".png"):
        output_file_path += ".png"
    mask = np.array(Image.open(mask_path))
    if len(mask.shape) == 2:
        mask = np.stack((mask,) * 3, axis=-1)  # convert grayscale to RGB

    # Calculate the aspect ratio of the mask
    mask_width, mask_height = mask.shape[1], mask.shape[0]
    mask_aspect_ratio = mask_width / mask_height

    # Adjust the width and height parameters to maintain the aspect ratio of the mask
    if width / height > mask_aspect_ratio:
        # The wordcloud will be constrained by height
        width = int(height * mask_aspect_ratio)
    else:
        # The wordcloud will be constrained by width
        height = int(width / mask_aspect_ratio)

    # Resize the mask image if it is smaller than the desired width and height
    mask_height, mask_width, _ = mask.shape
    if mask_width < width or mask_height < height:
        mask = cv2.resize(mask, (width, height))

    wc = WordCloud(mask=mask, width=width, height=height)
    wc.generate_from_frequencies(dict_word_count)
    # image_colors = ImageColorGenerator(mask)
    # wc.recolor(color_func=random_color)
    wc.to_file(output_file_path)


def generate_mask(image_path, mask_file_path=None):
    """
    Generates a mask image from an input image.

    Args:
        image_path: The path of the input image.
        mask_file_path: The path and file name to save the generated mask image. Default is 'mask.png'.

    Returns:
        The file path of the generated mask image.
    """
    if mask_file_path is None:
        mask_file_path = os.path.join(os.path.dirname(image_path), "mask.png")

    # Read the input image and convert it to grayscale
    image = np.array(Image.open(image_path).convert('L'))

    # Loop through different thresholds until the mask has at least 25% black pixels
    for threshold in range(256):
        # Threshold the image using the current threshold value
        binary = (image > threshold).astype(np.uint8) * 255

        # Calculate the percentage of black pixels in the binary image
        black_pixels_percent = np.count_nonzero(binary == 0) / binary.size

        # If the percentage of black pixels is at least 25%, save the binary image as the mask and return the file path
        if black_pixels_percent >= 0.25:
            Image.fromarray(binary).save(mask_file_path)
            return mask_file_path

    # If no threshold results in a mask with at least 25% black pixels, raise an error
    raise ValueError(
        "Could not generate a mask with at least 25% black pixels")

def folder(folder: Folder, bag_of_words_name = "00_bag_of_words", named_entities_name = "00_named_entities", bag_of_words_mask = "00_bag_of_words_mask", named_entities_name_mask = "00_named_entities_mask"):
    
    generate_wordcloud(
        folder.bag_of_words.get(), os.path.join(folder.path, bag_of_words_name))
    generate_wordcloud(
        folder.named_entities.get_frequencies(), os.path.join(folder.path, named_entities_name))
    mask_path = folder.get_image()
    if mask_path:
        mask = generate_mask(mask_path)
        generate_wordcloud_mask(
            folder.bag_of_words.get(), mask, os.path.join(folder.path, bag_of_words_mask))
        generate_wordcloud_mask(folder.named_entities.get_frequencies(
        ), mask, os.path.join(folder.path, named_entities_name_mask))


import time
import json
import warnings
import sys
import os
from os.path import isfile, join
from os import listdir
from datetime import datetime
try:
    import whisper
except ImportError:
    warnings.warn("Not using OpenAI whisper. If you want to use it, run 'pip install bnw_tools[whisper]'")


def benchmark(filename):
    old_stdout = sys.stdout
    with open("logfile.log", "a", encoding='utf-8') as log_file:
        sys.stdout = log_file

        checkpoint = time.time()
        print(str(time.time() - checkpoint) + '\n')

        model = whisper.load_model("tiny")
        result = model.transcribe(filename)
        with open('01_tiny.json'.format(1), 'w', encoding='utf-8') as file:
            json.dump(result, file, ensure_ascii=False, indent=4)

        checkpoint = time.time()
        print(str(time.time() - checkpoint) + '\n')

        model = whisper.load_model("base")
        result = model.transcribe(filename)
        with open('02_bas.json'.format(1), 'w', encoding='utf-8') as file:
            json.dump(result, file, ensure_ascii=False, indent=4)

        checkpoint = time.time()
        print(str(time.time() - checkpoint) + '\n')

        model = whisper.load_model("small")
        result = model.transcribe(filename)
        with open('03_small.json'.format(1), 'w', encoding='utf-8') as file:
            json.dump(result, file, ensure_ascii=False, indent=4)

        checkpoint = time.time()
        print(str(time.time() - checkpoint) + '\n')

        model = whisper.load_model("medium")
        result = model.transcribe(filename)
        with open('04_medium.json'.format(1), 'w', encoding='utf-8') as file:
            json.dump(result, file, ensure_ascii=False, indent=4)

        checkpoint = time.time()
        print(str(time.time() - checkpoint) + '\n')

        # if graphics card is strong enough:

        # model = whisper.load_model("large")
        # result = model.transcribe(filename)
        # with open('01_tiny.json'.format(1), 'w', encoding='utf-8') as file:
        #     json.dump(result, file, ensure_ascii=False, indent=4)

        # checkpoint = str(time.time())
        # print(checkpoint + '\n')

        # model = whisper.load_model("large-v2")
        # result = model.transcribe(filename)
        # json_object = json.dumps(result, indent=4)
        # with open("06_large-v2.json", "w", encoding='utf8') as outfile:
        #     outfile.write(json_object)

        # checkpoint = str(time.time())
        # print(checkpoint + '\n')
    sys.stdout = old_stdout


def grab(input_path, filename, output_path=None, device='cuda'):
    model = whisper.load_model("medium", device=device)

    result = model.transcribe(os.path.join(input_path, filename))
    if filename.endswith(".mp4"):
        newfilename_new = filename.replace('.mp4', '.json')
    elif filename.endswith(".mp3"):
        newfilename_new = filename.replace('.mp3', '.json')
    if output_path is None:
        output_path = input_path
    with open(os.path.join(output_path, newfilename_new), 'w', encoding='utf-8') as file:
        json.dump(result, file, ensure_ascii=False, indent=4)


def grab_list(input_path, files, device='cuda'):
    model = whisper.load_model("medium", device=device)
    queue = []
    for filename in files:
        queue.append(os.path.join(input_path, filename))
    results = model.transcribe()
    for idx, result in enumerate(results):
        if files[idx].endswith(".mp4"):
            newfilename_new = files[idx].replace('.mp4', '.json')
        elif files[idx].endswith(".mp3"):
            newfilename_new = files[idx].replace('.mp3', '.json')
        with open(os.path.join(input_path, newfilename_new), 'w', encoding='utf-8') as file:
            json.dump(result, file, ensure_ascii=False, indent=4)


def redo(filenames):
    file_path = os.getcwd() + '\\src\\redo.txt'
    with open(file_path, "r", encoding="utf8") as f:
        lines = f.readlines()
    files = [x.replace('stuck in no-punctuation mode: ', '') for x in lines]
    for file in files:
        file = file.replace('\n', '')
        # correct with files from redone
    return files


def redone(filename):
    file_path = os.getcwd() + '\\src\\redone.txt'
    with open(file_path, "r", encoding="utf8") as f:
        lines = f.readlines()
    files = [x.replace('stuck in no-punctuation mode: ', '') for x in lines]
    for file in files:
        file = file.replace('\n', '')
    return files


def extract_info(input_path, playlist_name=None):
    log_file = open("logfile.log", "a", encoding='utf-8')
    old_stdout = sys.stdout
    # todo: make this cleaner and load the name from central place
    audio_path = os.path.join(input_path, "00_audios")
    whisper_path = os.path.join(input_path, "01_whisper")
    with open("logfile.log", "a", encoding='utf-8') as log_file:
        sys.stdout = log_file
        try:
            filenames = [f for f in listdir(
                audio_path) if isfile(join(audio_path, f))]
            existing = [f for f in listdir(
                whisper_path) if isfile(join(whisper_path, f))]
        except:
            return
        print('\n-------------\n', flush=True)
        print(str(datetime.now()) + '\n', flush=True)
        last_skipped = ""
        skipped = 0
        # redo_filenames = redo(filenames)
        for filename in filenames:
            if '.mp4' in filename or '.mp3' in filename:
                checkpoint = time.time()
                if filename.endswith(".mp4"):
                    newfilename_new = filename.replace('.mp4', '.json')
                elif filename.endswith(".mp3"):
                    newfilename_new = filename.replace('.mp3', '.json')
                if newfilename_new not in existing:
                    if skipped > 0:
                        print('Skipping ' + str(skipped) +
                              ' until including ' + last_skipped + '.\n', flush=True)
                        skipped = 0
                    grab(audio_path, filename, output_path=whisper_path)
                    print(str(datetime.now()), flush=True)
                    print(filename + ': ' + str(round(time.time() -
                                                      checkpoint, 2)) + " Sek." + '\n', flush=True)
                # elif filename in redo_filenames:
                #     if skipped > 0:
                #         print('Skipping ' + str(skipped) + ' until including ' + last_skipped + '.\n', flush=True)
                #         skipped = 0
                #     print('Redoing ' + filename + '.\n', flush=True)
                #     grab(input_path, filename)
                #     print(str(datetime.now()), flush=True)
                #     print(filename + ': ' + str(round(time.time()-checkpoint, 2)) + " Sek." + '\n', flush=True)
                #     redone(filename)
                else:
                    last_skipped = filename
                    skipped += 1
    sys.stdout = old_stdout


def main(my_path=None):
    my_path = os.getcwd()
    data_path = os.path.dirname(my_path) + '\\data\\'
    playlist_names = [f for f in listdir(
        data_path) if not isfile(join(data_path, f))]
    playlist_names.remove('sample')
    for pl_n in playlist_names:
        data_pl_path = data_path + pl_n + '\\mp3'
        extract_info(data_pl_path, pl_n)


if __name__ == '__main__':
    main()

# file formats
# vtt
# srt

# txt
# json

# with open("Output.txt", "w") as text_file:
#     text_file.write(result + '\n')



#TODO: This file is deprecated and should be removed



import stanza
import publish.util_networkx as nx
import extract.util_pdf as util_pdf
import re
import os
# stanza.download('de')       # This downloads the English models for the neural pipeline
# This sets up a default neural pipeline in English
nlp = stanza.Pipeline('de')

#TODO: Check if this needs to be removed
try_pages = True
with open("input.txt", encoding="utf8") as f:
    text = f.read()

def connect_words(text, page_id, doc):
    data = []
    data_set = set()
    love = ["root", "obj", "csubj", "amod", "conj", "nsubj"]
    hate = ["punct", "det", "case", "cc"]

    for sent in doc.sentences:
        for word in sent.words:
            id = word.id
            lemma = word.lemma
            text = word.text
            if not re.search(r".*[a-zA-Z].*", text):
                continue
            head_id = word.head
            head = sent.words[word.head-1].text if word.head > 0 else "root"
            deprel = word.deprel
            if deprel in hate:
                continue
            head_lemma = "root"
            s_head_id = head_id
            distance_to_root = 0
            while s_head_id:
                distance_to_root += 1
                s_head_id = sent.words[s_head_id-1].head
            if distance_to_root > 0 and distance_to_root < 4:
                pre = sent.words[word.head-1]
                if pre.deprel not in hate:
                    head_lemma = pre.lemma
                    data_set.add(f'"{lemma}"')
                    data_set.add(f'"{head_lemma}"')
                    data.append([f'"{lemma}"', f'"{head_lemma}"'])
    filename = f"{page_id}_{pdf_name}.dot"
    # if page_id == len(text_pages)-1:
    #     filename == f"{pdf_name}_full.dot"
    directory = "output"
    if not os.path.exists(directory):
        os.mkdir(directory)
    with open(directory + filename, "w", encoding="utf8") as f:
        f.write("strict digraph  {\n")
        for lem in data_set:
            f.write(f"{lem}\n")
        for pair in data:
            f.write(f"{pair[0]} -> {pair[1]}\n")
        f.write("}\n")
    nx.dot_plot(filename, directory)

    # nx.graph_from_list(data)


def find_followers(leader, words):
    temp = {}
    for word in words:
        if word.head == leader:
            temp.update({word.id: find_followers(word.id, words)})
    return temp


def print_order(follower, words, ind=0):
    for leader in follower:
        pre = ""
        for i in range(ind):
            pre += "\t"
        print(pre + words[leader-1].text)
        print_order(follower[leader], words, ind+1)


def create_kg(text, page_id, doc):
    # data = []
    # data_set = set()
    # love = ["root", "obj", "csubj", "amod", "conj", "nsubj"]
    # hate = ["punct", "det", "case", "cc"]

    for sent in doc.sentences:
        subject = []
        predicate = []
        object = []

        for word in sent.words:
            if "obj" in word.deprel:
                object.append(word.id)
            elif "subj" in word.deprel:
                subject.append(word.id)
            elif "root" in word.deprel:
                predicate.append(word.id)

            # id = word.id
            # lemma = word.lemma
            # text = word.text
            # if not re.search(r".*[a-zA-Z].*", text):
            #     continue
            # head_id = word.head
            # head = sent.words[word.head-1].text if word.head > 0 else "root"
            # deprel = word.deprel
            # if deprel in hate:
            #     continue
            # head_lemma = "root"
            # s_head_id = head_id
            # distance_to_root = 0
            # while s_head_id:
            #     distance_to_root += 1
            #     s_head_id = sent.words[s_head_id-1].head
            # if distance_to_root > 0 and distance_to_root < 4:
            #     pre = sent.words[word.head-1]
            #     if pre.deprel not in hate:
            #         head_lemma = pre.lemma
            #         data_set.add(f'"{lemma}"')
            #         data_set.add(f'"{head_lemma}"')
            #         data.append([f'"{lemma}"', f'"{head_lemma}"'])
        if len(subject) > 1:
            pass
            # handle somehow
        if len(object) > 1:
            pass
            # handle somehow
        if len(predicate) > 1:
            pass
            # handle somehow
        subject = subject[0]
        object = object[0]
        predicate = predicate[0]
        leader = -1
        for word in sent.words:
            if word.head == 0:
                leader = word.id
                break
        follower = {
            leader: {}
        }
        follower[leader].update(find_followers(leader, sent.words))
        print_order(follower, sent.words)
        pass

    filename = f"{page_id}_{pdf_name}.dot"
    # if page_id == len(text_pages)-1:
    #     filename == f"{pdf_name}_full.dot"
    directory = "output"
    if not os.path.exists(directory):
        os.mkdir(directory)
    with open(directory + filename, "w", encoding="utf8") as f:
        f.write("strict digraph  {\n")
        for lem in data_set:
            f.write(f"{lem}\n")
        for pair in data:
            f.write(f"{pair[0]} -> {pair[1]}\n")
        f.write("}\n")
    nx.dot_plot(filename, directory)

    # nx.graph_from_list(data)


def handle_text(text, page_id=0):
    doc = nlp(text)
    print(str(page_id))
    # doc.sentences[0].print_dependencies()

    print(*[f'id: {word.id}\tword: {word.text}\thead id: {word.head}\thead: {sent.words[word.head-1].text if word.head > 0 else "root"}\tdeprel: {word.deprel}' for sent in doc.sentences for word in sent.words], sep='\n')
    # connect_words(text, page_id, doc)
    create_kg(text, page_id, doc)


if try_pages:
    text_pages = re.split(r" \d{1,2}\n\d{1,2} ", text)
    for page_id, text in enumerate(text_pages):
        handle_text(text, page_id)
else:
    handle_text(text)



import yaml
import os
import sys
import bnw_tools
# 
# from importlib.util import find_spec
# bnw_tools_spec = find_spec('bnw_tools')

# if bnw_tools_spec is None:
    # sys.path.append(os.getcwd())
    # import bnw_tools

# import bnw_tools

# todo: make this import cleaner
# sys.path.append(os.getcwd())
# sys.path.append(os.path.join(os.getcwd(), "bnw_tools"))

from bnw_tools.extract import util_pytube
from bnw_tools.publish.Obsidian import nlped_whispered_folder
from bnw_tools.publish import util_wordcloud
from bnw_tools.extract.nlp import util_nlp
from bnw_tools.extract import util_whisper


queue = [
    # {"folder": "D:/workspace/Zotero/pipelines",
    #  "language": "en"},
    # {"folder": "D:/workspace/Zotero/SE2A-B4-2",
    #  "language": "en"},
    # "language": "de"},
    {"folder": "D:/workspace/raw (MP3)/HoP",
     "image": "D:/workspace/raw (MP3)/HoP/mask/ab67656300005f1fcabcff3dcfa8fbd5f5b0fa04.jpeg",
     "language": "en"}
]


channels = {
    # "UCwBT6zl54_cP0mOrVPFH-qg": "BorgNetzWerk",
    # "UCoYvhMd-3LbORHvBVkC0hSw": "orkg7258",
    # "cinematherapysolutions": "CinemaTherapyShow",
    # "UCyHDQ5C6z1NDmJ4g6SerW8g": "maiLab",
    # "inanutshell": "kurzgesagt",
    # "sciphi635": "sciphi635",
    # "unbubble": "unbubble",
    # "georgiadow": "GeorgiaDow",
    # "LegalEagle": "LegalEagle",
    # "coldmirror": "coldmirror",
    # "UCPtUzxTfdaxAmr4ie9bXZVA": "MathebyDanielJung",
    # "ROCKETBEANSTV": "ROCKETBEANSTV",
}

YT_root = "D:/workspace/YouTube"

for channel_id, channel_name in channels.items():
    entry = {"folder": os.path.join(YT_root, channel_name),
             "channel_id": channel_id,
             "channel_name": channel_name}
    queue.append(entry)

# Extract
# Transcribe (Whisper)
nlptools = util_nlp.NLPTools()
for do in queue:
    folder_path = do["folder"]
    language = None
    if language in do:
        language = do["language"]
    old_stdout = sys.stdout
    # todo: make this clearer
    parent = os.path.basename(os.path.dirname(folder_path))
    if parent == "raw (MP3)":
        util_whisper.extract_info(folder_path)
    elif parent == "Zotero":
        # todo: make this productive
        pass
    elif parent == "YouTube":
        channel_id = do["channel_id"]
        channel_name = do["channel_name"]
        res = util_pytube.do_channels(YT_root, channel_id, channel_name)
        if res and "language" in res:
            language = res["language"]
        else:
            language = "en"
        pass
        # todo: make this productive

    sys.stdout = old_stdout
    # NLP (Flair and SpaCy)
    folder = util_nlp.Folder(
        folder_path, nlptools=nlptools, language=language)

    # Wordcloud
    if folder.media_resources:
        util_wordcloud.generate_wordcloud(
            folder.bag_of_words.get(), os.path.join(folder_path, "00_bag_of_words"))
        util_wordcloud.generate_wordcloud(
            folder.named_entities.get_frequencies(), os.path.join(folder_path, "00_named_entities"))
        mask_path = folder.get_image()
        if mask_path:
            mask = util_wordcloud.generate_mask(mask_path)
            util_wordcloud.generate_wordcloud_mask(
                folder.bag_of_words.get(), mask, os.path.join(folder_path, "00_bag_of_words_mask"))
            util_wordcloud.generate_wordcloud_mask(folder.named_entities.get_frequencies(
            ), mask, os.path.join(folder_path, "00_named_entities_mask"))

        # Obsidian
        nlped_whispered_folder.folder(folder, force=True)



import wiki.barlow as barlow

def write_individual_wiki_page(p_n, ID, episode_info):
    if p_n == 'BMZ':
        return barlow.write_individual_wiki_page_BMZ(ID, episode_info)

def write_episodes(p_n, episodes_info):
    if p_n == 'BMZ':
        return barlow.write_mainpage(episodes_info)

# acro = barlow.acro


import argparse
import os
import json
import spacy
from spacy.cli.download import download
from flair.data import Sentence, Token
from flair.models import SequenceTagger
from flair.embeddings import WordEmbeddings, DocumentPoolEmbeddings
from collections import defaultdict
from typing import List, Dict, Union, Optional
import inspect
import numpy as np
import math
import re
import urllib.request
from mutagen.easyid3 import EasyID3
import importlib
from termcolor import colored

from ...review import similarity
from ...core import helper
from ...publish import util_wordcloud
from ...publish.Obsidian import nlped_whispered_folder
from ...extract import util_pdf
from ...extract import util_zotero
from ...extract import util_pytube
from ...extract.nlp.my_stop_words import MY_STOP_WORDS


def get_json_data(path):
    if os.path.exists(path) and os.path.splitext(path)[1] == ".json":
        try:
            with open(path, 'r', encoding='utf8') as f:
                return json.load(f)
        except Exception as e:
            print(f"Could not load {path}: {str(e)}")
            return None


def dump_json(path, output_dict):
    with open(path, 'w', encoding='utf-8') as output_file:
        json.dump(output_dict, output_file, indent=4, ensure_ascii=False)


def data_to_attributes(instance, data, do_named_entities=True):
    if not isinstance(data, dict):
        return None
    words = None
    stops = None
    for key, value in data.items():
        if key == "bag_of_words":
            words = value
            continue
        elif key == "stop_words":
            stops = value
            continue
        if hasattr(instance, key):
            if key == "named_entities":
                if do_named_entities:
                    setattr(instance, key, NamedEntities(value))
                else:
                    setattr(instance, key, NamedEntities())
            else:
                setattr(instance, key, value)
    if words or stops:
        setattr(instance, "bag_of_words", BagOfWords(words=words, stops=stops))


class NLPTools:
    naming_conventions = {
        "en": {"spacy": "en_core_web_sm", "SequenceTagger": "ner", "WordEmbeddings": "glove"},
        "de": {"spacy": "de_core_news_sm", "SequenceTagger": "flair/ner-german", "WordEmbeddings": "de"},
        # add more languages and their naming conventions as needed
    }

    def __init__(self):
        self.STOP_WORDS = {}
        self.spacy = {}
        self.WordEmbeddings = {}
        self.DocumentPoolEmbeddings = {}
        self.SequenceTagger = {}

    def get_STOP_WORDS(self, language):
        if language in self.STOP_WORDS:
            return self.STOP_WORDS[language]
        stop_words_module = importlib.import_module(
            f"spacy.lang.{language}.stop_words")
        self.STOP_WORDS[language] = MY_STOP_WORDS[language].union(
            stop_words_module.STOP_WORDS)
        # TODO: Idea to allways include the english stop-words, since sometimes other languages use english as well
        return self.STOP_WORDS[language]

    def get_spacy(self, language):
        if language in self.spacy:
            return self.spacy[language]
        try:
            nlp_model = self.naming_conventions.get(
                language, {}).get("spacy", f"{language}_core_news_sm")
            try:
                self.spacy[language] = spacy.load(nlp_model)
            except OSError:
                download(model=nlp_model)
                self.spacy[language] = spacy.load(nlp_model)

            stop_words = self.get_STOP_WORDS(language)
            self.spacy[language].Defaults.stop_words |= set(stop_words)

            for word in self.get_STOP_WORDS(language):
                # todo: this does not seem to be enoguh, words with the same lemma get through
                # thought: this may be intended and will be left like this for now.
                for w in (word, word[0].capitalize(), word.upper()):
                    lex = self.spacy[language].vocab[w]
                    lex.is_stop = True
            return self.spacy[language]
        except Exception as e:
            print(f"Error loading spacy: {str(e)}")
            return None

    def get_WordEmbeddings(self, language):
        if language in self.WordEmbeddings:
            return self.WordEmbeddings[language]
        try:
            word_embeddings_model = self.naming_conventions.get(
                language, {}).get("WordEmbeddings", language)
            self.WordEmbeddings[language] = WordEmbeddings(
                word_embeddings_model)
            return self.WordEmbeddings[language]
        except Exception as e:
            print(f"Error loading WordEmbeddings for {language}: {str(e)}")
            return None

    def get_DocumentPoolEmbeddings(self, language):
        if language in self.DocumentPoolEmbeddings:
            return self.DocumentPoolEmbeddings[language]
        try:
            self.DocumentPoolEmbeddings[language] = DocumentPoolEmbeddings(
                [self.get_WordEmbeddings(language)])
            return self.DocumentPoolEmbeddings[language]
        except Exception as e:
            print(
                f"Error loading DocumentPoolEmbeddings for {language}: {str(e)}")
            return None

    def get_SequenceTagger(self, language):
        if language in self.SequenceTagger:
            return self.SequenceTagger[language]
        try:
            sequence_tagger_model = self.naming_conventions.get(
                language, {}).get("SequenceTagger", f"flair/ner-{language.lower()}")
            self.SequenceTagger[language] = SequenceTagger.load(
                sequence_tagger_model)
            return self.SequenceTagger[language]
        except Exception as e:
            print(f"Error loading SequenceTagger for {language}: {str(e)}")
            return None


class BagOfWords:
    def __init__(self, words: Dict[str, int] = None, stops: Dict[str, int] = None, tf: Dict[str, float] = None, tf_idf: Dict[str, float] = None, text=None, nlptools=None, language=None):
        self.words = words or defaultdict(int)
        self.stops = stops or defaultdict(int)
        self.tf = tf or {}
        self.tf_idf = tf_idf or {}

        if self.words and language:
            self.update_stops(language, nlptools)

        if text:
            self.from_nlp_text(text, language, nlptools)

    def __len__(self):
        return len(self.words)

    def calc_tf(self):
        total = sum(self.words.values())
        self.tf = {word: count / total for word, count in self.words.items()}
        self.sort(do_words=False, do_tf=True)
        return self.tf

    def calc_tf_idf(self, idf: Dict[str, float]):
        self.tf_idf = {word: frequency * idf[word]
                       for word, frequency in self.get_tf().items()}
        self.sort(do_words=False, do_tf_idf=True)
        return self.tf_idf

    def get_tf(self):
        if not self.tf:
            return self.calc_tf()
        return self.tf

    def get_tf_idf(self, idf: Dict[str, float] = None):
        if not self.tf_idf and idf:
            return self.calc_tf_idf(idf)
        return self.tf_idf

    def sort(self, do_words=True, do_tf=False, do_tf_idf=False):
        if do_words:
            if len(self.words):
                # self.words = {k: v for k, v in sorted(self.words.items(),
                #   key=lambda x: x[1], reverse=True)}
                self.words = dict(
                    sorted(self.words.items(), key=lambda x: (-x[1], x[0])))

            if len(self.stops):
                # self.stops = {k: v for k, v in sorted(self.stops.items(),
                #   key=lambda x: x[1], reverse=True)}
                self.stops = dict(
                    sorted(self.stops.items(), key=lambda x: (-x[1], x[0])))
        if do_tf:
            if len(self.tf):
                self.tf = dict(
                    sorted(self.tf.items(), key=lambda x: (-x[1], x[0])))
        if do_tf_idf:
            if len(self.tf_idf):
                self.tf_idf = dict(
                    sorted(self.tf_idf.items(), key=lambda x: (-x[1], x[0])))

    def get(self):
        return self.words

    def add(self, other):
        for word, count in other.words.items():
            if word in self.words:
                self.words[word] += count
            else:
                self.words[word] = count

    def from_nlp_text(self, nlp_text, language=None, nlpTools=None):
        for token in nlp_text:
            if token.is_alpha:
                word = token.lemma_.lower()
                if token.is_stop:
                    self.stops[word] += 1
                else:
                    self.words[word] += 1
        self.update_stops(language=language, nlptools=nlpTools)
        self.sort()

    def update_stops(self, language, nlptools: NLPTools = None):
        def special_stop_rules(key, vocab):
            if len(key) < 2:
                return True
            if not key.isalpha() and key not in vocab:
                if len(re.findall('[a-zA-Z\d]', key)) < 2:
                    return True
            # way to harsh, not every word is in (that language's) vocab
            # if key not in vocab:
            #     return True
            return False

        nlptools = nlptools or NLPTools()
        to_stop = []
        to_bag = []
        vocab = nlptools.get_spacy(language).vocab
        for key in self.words.keys():
            if key in vocab and vocab[key].is_stop:
                to_stop.append(key)
            elif special_stop_rules(key, vocab):
                to_stop.append(key)
        for key in self.stops.keys():
            if key in vocab and not vocab[key].is_stop:
                to_bag.append(key)
        for key in to_stop:
            self.stops[key] = self.words.pop(key)
        for key in to_bag:
            self.words[key] = self.stops.pop(key)
        self.sort()
        return len(to_bag) + len(to_stop)

    # bag_of_words = {k: v for k, v in sorted(bag_of_words.items(), key=lambda item: item[1], reverse=True)}


class NamedEntities:
    def __init__(self, entities: Dict[str, List[str]] = None, text=None,):
        self.entities = entities or {}

        if text:
            self.from_nlp_text(text)

    def __len__(self):
        return len(self.entities)

    def sort(self):
        if not self.entities:
            return
        val = list(self.entities.values())[0]
        if isinstance(val, list):
            self.entities = {k: v for k, v in sorted(
                self.entities.items(), key=lambda item: len(item[1]), reverse=True)}
        if isinstance(val, int):
            self.entities = {k: v for k, v in sorted(
                self.entities.items(), key=lambda item: item[1], reverse=True)}

    def add(self, other, path=None):
        for key, value in other.entities.items():
            if path:
                path = os.path.basename(path)
                value = [f"{path}: {x}" for x in value]
            if key in self.entities:
                self.entities[key] += value
            else:
                self.entities[key] = value

    def update(self):
        move_to = {}
        for key in self.entities.keys():
            if key.lower() == key:
                for other_key in self.entities.keys():
                    if other_key != key and other_key.lower() == key:
                        move_to[key] = other_key
                        break
        for source, dest in move_to.items():
            if isinstance(self.entities[dest], list):
                self.entities[dest] += self.entities[source]
            else:
                self.entities[dest].update(self.entities[source])
            del self.entities[source]

    def get(self):
        return self.entities

    def get_frequencies(self):
        return {k: len(v) for k, v in self.entities.items()}

    def from_nlp_text(self, sentence):
        for entity in sentence.get_spans('ner'):
            rep = f"Span[{entity.start_position}:{entity.end_position}]: {entity.tag} ({round(entity.score, 2)})"
            if entity.text in self.entities:
                self.entities[entity.text].append(rep)
            else:
                self.entities[entity.text] = [rep]
        self.sort()

# File classes


class Transcript:
    def __init__(self, path=None, text=None, segments=None, language=None):
        """
        A class representing a transcript.

        Args:
            text (str): Text content of the transcript.
            path (str): Path to the transcript file.

        Attributes:
            text (str): Text content of the transcript.
            path (str): Path to the transcript file.
        """
        self.path: str = path
        self.text: str = text
        # todo: eventually have a "clean_text" variable
        self.language: str = language
        self.segments: Dict[str, str] = segments
        if path:
            self.from_file()

    def get_manageable_text_sizes(self):
        text = self.text
        pieces = []
        jump = 50000

        if len(text) < jump * 2:
            pieces = [text]
        else:
            # Break text into smaller chunks so we have enough memory to solve it
            # FIXME: doesn't properly break
            candidates = ["\n\n", "\n", ".", "!", "?", ",", " "]

            begin = 0
            end = jump
            while end < len(text):
                next_stop = 0
                for fi in candidates:
                    pos = text[begin + jump:end + jump].find(fi)
                    if pos >= 0 and pos <= jump * 2:
                        next_stop = pos
                        break
                end = begin + jump + next_stop + 1
                pieces.append(text[begin:end])
                begin = end
            pieces.append(text[begin:])
        return pieces

    def from_file(self, path=None):
        """
        Load transcript from the given path.

        Args:
            path (str): Path to the transcript file.
        """
        if path is None:
            path = self.path
        else:
            self.path = path
        if not path:
            print("need path to load Transcript")
            return None
        data = get_json_data(path)
        data_to_attributes(self, data)

    def iscomplete(self):
        if self.path and self.text:
            return True
        return False

    def complete(self, path=None):
        if not self.iscomplete() and path:
            self.from_file(path)
        else:
            print("missing path to load transcript")
        return self.iscomplete()

    def get_dict(self):
        return {
            'text': self.text,
            'segments': self.segments,
            'language': self.language
        }

    def get(self, attr: str):
        if attr == "dict":
            return self.get_dict()
        elif hasattr(self, attr):
            return getattr(self, attr)

    def clean(self):
        if self.text and self.text.startswith("1\n00:00:"):
            # is vtt transcript
            # make sure empty vtt lines are not "truly" emtpy
            self.text = self.text.replace("\n\n\n", "\n \n\n")
            segments = self.text.split("\n\n")
            sentences = [s.split("\n")[2] for s in segments]
            # todo: detect when speaker is given like "Speaker a: ..."
            speaker_splitter = [":", "]"]
            for s_c in speaker_splitter:
                openings = defaultdict(int)
                for sentence in sentences:
                    pieces = sentence.split(s_c)
                    if len(pieces) > 1:
                        openings[pieces[0]] += 1

                speakers = False
                # see if there are regular speakers:
                for opening, count in openings.items():
                    if count > 5:
                        speakers = True
                if speakers:
                    for idx, sentence in enumerate(sentences):
                        pieces = sentence.split(s_c)
                        if len(pieces) > 1:
                            pieces.pop(0)
                        sentences[idx] = s_c.join(pieces)
            sentences = [s for s in sentences if s.count(" ") != len(s)]
            self.text = " ".join(sentences)

        elif "\n" in self.text:
            self.text = self.text.replace("-\n", "")
            self.text = self.text.replace("\n", " ")


class Audio:
    def __init__(self, path=None, metadata=None):
        self.path = path
        self.metadata = metadata
        if self.path:
            try:
                self.metadata = EasyID3(path)
            except:
                pass

    def get(self, arg=None):
        if arg == "path":
            return self.path
        elif arg == "title":
            return os.path.splitext(os.path.basename(self.path))[0]
        elif arg:
            return self.metadata.get(arg)
        else:
            return self.metadata

    def get_dict(self):
        return self.metadata.items()


class NLPFeatureAnalysis:
    def __init__(self, path=None, bag_of_words=None, named_entities=None, nlptools: NLPTools = None, transcript: Transcript = None, do_named_entities=True):
        """
        A class representing NLP feature analysis.

        Args:
            bag_of_words (dict): A dictionary representing bag of words.
            named_entities (list): A list of named entities.
            tf_idf (dict): A dictionary representing tf-idf values.

        Attributes:
            bag_of_words (dict): A dictionary representing bag of words.
            named_entities (list): A list of named entities.
            tf_idf (dict): A dictionary representing tf-idf values.
        """
        self.path: str = path
        self.bag_of_words: BagOfWords = bag_of_words or BagOfWords()
        self.named_entities: NamedEntities = named_entities or NamedEntities()
        if path:
            self.from_file(path, do_named_entities=do_named_entities)
        if transcript and transcript.text:
            self.complete(transcript, nlptools,
                          do_named_entities=do_named_entities)

    def from_file(self, path: str = None, do_named_entities: bool=True) -> None:
        """
        Load transcript from the given path.

        Args:
            path (str): Path to the transcript file.
        """
        if path is None:
            path = self.path
        else:
            self.path = path
        data = get_json_data(path)
        data_to_attributes(self, data, do_named_entities)

    def iscomplete(self) -> bool:
        if self.bag_of_words.get() and self.named_entities.get():
            return True
        return False

    def complete(self, transcript: Transcript, nlptools: NLPTools = None, do_named_entities: bool = True) -> bool:
        """Processes a JSON file containing text and returns a Document object.

        Args:
            json_path (str): The path to the input JSON file.

        Returns:
            Document: The Document object representing the input text.
        """
        nlptools = nlptools or NLPTools()

        changes = False
        if not self.iscomplete():
            changes = True
        if not transcript or not transcript.text:
            print("No Text in: " + self.path)
            return False
        transcript.clean()
        if do_named_entities:
            self.fill_named_entities(transcript, nlptools)
        self.fill_bag_of_words(transcript, nlptools)

        # create a Document object and return it
        entity_change = self.named_entities.update()
        BagOfWords_change = self.bag_of_words.update_stops(
            transcript.language, nlptools)
        if changes or BagOfWords_change or entity_change:
            self.save(should_have_nen=do_named_entities)
        return self.iscomplete()

    def fill_bag_of_words(self, transcript: Transcript, nlptools: NLPTools = None) -> Union[BagOfWords, bool]:
        if not self.bag_of_words.get():
            nlptools = nlptools or NLPTools()
            if not transcript.iscomplete():
                print("You need a text in a known language to complete " + self.path)
                return False
            # create a Sentence object from the text

            bow = BagOfWords()

            for text in transcript.get_manageable_text_sizes():
                doc = nlptools.get_spacy(transcript.language)(text)

                bow.add(BagOfWords(
                    text=doc, nlptools=nlptools, language=transcript.language))

            # get the bag of words and non-stop-words
            self.bag_of_words = bow
            return self.bag_of_words
        else:
            self.bag_of_words.update_stops(
                language=transcript.language, nlptools=nlptools)
            return False

    def fill_named_entities(self, transcript: Transcript, nlptools: NLPTools = None) -> Union[NamedEntities, bool]:
        if not self.named_entities.get():
            nlptools = nlptools or NLPTools()
            if not transcript.iscomplete():
                print("You need a text in a known language to complete " + self.path)
                return False

            nen = NamedEntities()

            for text in transcript.get_manageable_text_sizes():
                sentence = Sentence(text)

                nlptools.get_DocumentPoolEmbeddings(
                    transcript.language).embed(sentence)

                # get the named entities
                # todo: Worked
                nlptools.get_SequenceTagger(
                    transcript.language).predict(sentence)
                nen.add(NamedEntities(text=sentence))
            self.named_entities = nen
            return self.named_entities
        else:
            return False

    def get_dict(self) -> dict:
        # Todo: consider writing 'tf' to file as well.
        return {
            'bag_of_words': self.bag_of_words.words,
            'tf_idf': self.bag_of_words.get_tf_idf(),
            'named_entities': self.named_entities.get(),
            'stop_words': self.bag_of_words.stops
        }

    def save(self, path: Optional[str] = None, should_have_nen: bool = True) -> None:
        if path is None:
            path = self.path
        output_dict = self.get_dict()

        if len(self.bag_of_words.words) == 0:
            print(f"No {colored('words', 'red')} in file: " + path)
        elif len(self.named_entities) == 0 and should_have_nen:
            print(f"No {colored('named entities', 'red')} in file: " + path)
        dump_json(path, output_dict)


class ResearchQuestion:
    def __init__(self, title: str = None, path: str = None, keywords: dict[str, int] = None, queries: dict[str, str] = None):
        self.path: str = None
        self.title: str = title
        self.keywords: dict[str, int] = keywords
        self.queries: dict[str, str] = queries
        self.tag: str = None

    def from_file(self, path: str = None):
        if path is None:
            path = self.path
        else:
            self.path = path
        with open(path, "r", encoding="utf-8") as f:
            lines = f.readlines()
            if not lines or not lines[0].startswith("RQ::"):
                return None
        while "\n" in lines:
            lines.remove("\n")
        while "```\n" in lines:
            lines.remove("```\n")
        self.title = lines.pop(0).strip().split("::")[1].strip()

        self.tag = helper.get_clean_title(
            self.title, obsidian=True).replace(" ", "_")
        tag_found = False

        self.keywords = {}
        self.queries = {}
        in_keywords = False
        in_queries = False
        current_name = ""
        current_group = {}
        current_query = ""

        for line in lines:
            if self.tag in line:
                tag_found = True
            if line.startswith("## "):
                if in_keywords:
                    if current_name:
                        self.keywords[current_name] = current_group
                        current_name = ""
                        current_group = {}
                    in_keywords = False
                    in_queries = True
                elif in_queries:
                    if current_name:
                        self.queries[current_name] = current_query
                        current_name = ""
                        current_query = ""
                    in_queries = False
                    # whatever_comes_next = True
                else:
                    in_keywords = True
            elif in_keywords:
                if line.startswith("### "):

                    if current_name:
                        self.keywords[current_name] = current_group
                        current_name = ""
                        current_group = {}
                    current_name = line.strip().split(
                        "::")[0].split("### ")[1].lower()
                else:
                    keyword, weight = line.strip().split("::")
                    current_group[keyword.lower().strip()] = int(
                        weight.strip())
            elif in_queries:
                if line.startswith("### "):
                    if current_name:
                        self.queries[current_name] = current_query
                        current_name = ""
                        current_query = ""
                    current_name = line.strip().split("::")[0].split("### ")[1]
                else:
                    current_query += line.strip()

        if not tag_found:
            with open(path, "a", encoding="utf-8") as f:
                f.write(f"""
# Files
```dataview
Table
{self.tag} as "Fit"
from "03_notes"
SORT {self.tag} DESC
LIMIT 100
```
""")
        if current_name:
            self.queries[current_name] = current_query
            current_name = ""
            current_query = ""
        # creation successful
        return True

    def get_total_note(research_questions):
        pieces = []
        tags = []
        for i, key in enumerate(research_questions):
            tags.append(key.tag)
            pieces.append(f"{key.tag} as RQ{i+1}")
        rq_sum = " + ".join(tags)
        pieces.append(f"round({rq_sum},3) as \"Sum\"")
        total_cols = ", ".join(pieces)
        total_sort = f"SORT {rq_sum} desc"

        note = f"""## Research Questions
```dataview
Table
RQ as "RQ"
where RQ
SORT file.name
```

## Done
```dataview
Table
{total_cols}
From "03_notes"
{total_sort}
WHERE !contains(file.tasks.completed,false)
LIMIT 100
```

## Modified
```dataview
Table
{total_cols}
From "03_notes"
{total_sort}
WHERE contains(file.tasks.completed,true) AND contains(file.tasks.completed,false)
LIMIT 100
```

## Best Candidates
```dataview
Table
{total_cols}
From "03_notes"
{total_sort}
WHERE contains(file.tasks.completed,false)
LIMIT 100
```
"""
        return note


class MediaResource:
    """
    A class representing a media resource.

    Args:
        audio_file (Audiofile): An instance of Audiofile.
        transcript (Transcript): An instance of Transcript.
        nlp_analysis (NLPFeatureAnalysis): An instance of NLPFeatureAnalysis.

    Attributes:
        audio_file (Audiofile): An instance of Audiofile.
        transcript (Transcript): An instance of Transcript.
        nlp_analysis (NLPFeatureAnalysis): An instance of NLPFeatureAnalysis.
    """

    def __init__(self, nlptools: NLPTools = None):
        self.audio_file: Audio = None
        self.transcript: Transcript = None

        self.pdf: util_pdf.PDFDocument = None

        # todo: make this an actual class
        self.image: str = None
        self.info: util_pytube.YouTubeInfo = None

        self.nlp_analysis: NLPFeatureAnalysis = None

        self.obsidian_note: nlped_whispered_folder.ObsidianNote = None

        self.basename: str = None
        self.original_name: str = None

    def add_audio(self, path: str = None, file: Audio = None):
        if file:
            self.audio_file = file
        elif path:
            self.audio_file = Audio(path)
        self.search_original_name()

    def add_transcript(self, path: str = None, file: Transcript = None):
        if file:
            self.transcript = file
        elif path:
            self.transcript = Transcript(path)

    def add_info(self, path: str = None, file: util_pytube.YouTubeInfo = None):
        if file:
            self.info = file
        elif path:
            self.info = util_pytube.YouTubeInfo(path=path)

    def add_pdf(self, path: str = None, file: util_pdf.PDFDocument = None):
        if file:
            self.pdf = file
        elif path:
            self.pdf = util_pdf.PDFDocument(path)
        self.search_original_name()

    def add_NLPFeatureAnalysis(self, path: str = None, file: NLPFeatureAnalysis = None, nlptools: NLPTools = None, do_named_entities=True):
        if file:
            self.nlp_analysis = file
        elif path:
            if self.pdf:
                text_container = self.pdf
            elif self.transcript:
                text_container = self.transcript
            else:
                print(
                    f"Lacking text basis, can't add NLPFeatureAnalysis for file {self.basename}")
                return None
            self.nlp_analysis = NLPFeatureAnalysis(
                path, nlptools=nlptools, transcript=text_container, do_named_entities=do_named_entities)

    def add_image(self, path: str = None):
        if path:
            self.image = path

    def add_obsidian_note(self, path: str = None, file: nlped_whispered_folder.ObsidianNote = None):
        if file:
            self.obsidian_note = file
        elif path:
            self.obsidian_note = nlped_whispered_folder.ObsidianNote(path)
        else:
            self.obsidian_note = nlped_whispered_folder.ObsidianNote()

    def search_original_name(self, force=False):
        if self.original_name and not force:
            return self.original_name
        candidates = []
        if self.pdf:
            candidates.append(self.pdf.get("title"))
        if self.audio_file:
            candidates.append(self.audio_file.get("title"))
        if candidates:
            self.original_name = candidates[0]
            return self.original_name

    def iscomplete(self):
        return self.transcript.iscomplete() and self.nlp_analysis.iscomplete()

    def complete(self, nlptools: NLPTools = None, do_named_entities=True):
        """Processes a JSON file containing text and returns a Document object.

        Args:
            json_path (str): The path to the input JSON file.

        Returns:
            Document: The Document object representing the input text.
        """
        nlptools = nlptools or NLPTools()

        self.transcript.complete()
        self.nlp_analysis.complete(
            nlptools, do_named_entities=do_named_entities)

    def get_dict(self):
        # TODO: check if this throws errors, else revert
        # return dict(self.audio_file.get_dict()) | self.transcript.get_dict() | self.nlp_analysis.get_dict()
        result_dict = {}
        for key in self.__dict__:
            attr = getattr(self, key)
            try:
                # todo: remove all get_dicts
                if hasattr(attr, 'get_dict'):
                    result_dict |= attr.get_dict()
                elif hasattr(attr, 'get') and callable(attr.get):
                    result_dict |= attr.get('dict')
            except:
                continue
        return result_dict

    def get(self, parameter):
        if parameter == 'text':
            if self.pdf is not None:
                return self.pdf.get("text")
            elif self.transcript is not None:
                return self.transcript.get("text")
            else:
                return None
        elif parameter == 'url':
            return self.url
        elif parameter == 'dict':
            return self.get_dict()
        elif parameter == 'basename':
            if self.basename:
                return self.basename
            else:
                path = None
                if self.nlp_analysis:
                    path = self.nlp_analysis.path
                elif self.pdf:
                    path = self.info.pdf
                elif self.audio_file:
                    path = self.info.audio_file
                elif self.info:
                    path = self.info.path
                self.basename = os.path.splitext(os.path.basename(path))[0]
                return self.basename
        elif parameter == 'type':
            return self.__class__.__name__
        else:
            raise ValueError(f"Invalid parameter '{parameter}'")


class Subfolder:
    folder_formats = {
        "blank": [],
        # 00 Input
        "00_PDFs": [".pdf"],
        "00_audios": [".mp3", ".mp4"],
        "00_images": [".png", ".jpg", ".jpeg"],
        "00_infos": [".json"],
        "00_RQs": [".md", ".json"],
        # 01 Extract
        "01_whisper": [".json"],  # todo: vtt
        # 02 Analyse
        # todo: renaming from _processed
        "02_nlp": [".json", "_processed.json"],
        # 03 Publish
        "03_notes": [".md"],
    }

    def __init__(self, folder_path, type="blank"):
        self.path = os.path.join(folder_path, type)
        os.makedirs(self.path, exist_ok=True)
        self.type = type
        self.endings = self.folder_formats[type]
        if type == "00_PDFs":
            self.BibResources = util_zotero.BibResources(folder_path)
        if type == "00_RQs":
            self.research_questions: list[ResearchQuestion] = None
            self.get("research_questions")
            
    def lookfor(self, filename):
        path_found = []
        path_suggested = []
        basename = os.path.basename(filename)
        stem = os.path.splitext(basename)[0]
        for ending in self.endings:
            path = os.path.join(self.path, stem + ending)
            if os.path.exists(path):
                path_found.append(path)
            else:
                path_suggested.append(path)
        best_fit = path_found[0] if path_found else path_suggested[0] if path_suggested else None
        return best_fit
        # return found, not_found

    def get(self, arg: str = None):
        if arg.lower() == "rqs" or arg.lower() == "research_questions":
            if self.research_questions:
                return self.research_questions
            files = self.get("files")
            if not files:
                return None
            else:
                if not self.research_questions:
                    self.research_questions: list[ResearchQuestion] = []
                for file in self.get("files"):
                    rq = ResearchQuestion()
                    if rq.from_file(os.path.join(self.path, file)):
                        self.research_questions.append(rq)
                with open(os.path.join(self.path, file), "w", encoding="utf8") as f:
                    f.write(ResearchQuestion.get_total_note(
                        self.research_questions))
                return self.research_questions
        elif arg == "files":
            return os.listdir(self.path)


class Folder:
    """A class to represent a folder containing multiple Document instances.

    Attributes:
        documents (List[Document]): A list of Document instances contained in the folder.
        stop_words (defaultdict(int)): A defaultdict with stop words as keys and their total count across
            all documents as values.
        bag_of_words (defaultdict(int)): A defaultdict with words as keys and their total count across all
            documents as values.
        named_entities (Dict[str, List[str]]): A dictionary with named entities as keys and their occurrences in
            the documents as values.
        paths (Dict[str, str]): A dictionary of paths to subfolders.

    Methods:
        add_document(document: Document) -> None: Adds a Document instance to the folder.
        populate() -> None: Populates the stop_words, bag_of_words and named_entities attributes by aggregating
            the counts and occurrences of the respective attributes in all documents.
        save(output_path: str) -> None: Saves the Folder instance to a JSON file at the specified output path.
    """

    # Define paths as a class-level property

    def __init__(self, folder_path, nlptools=None, language=None, media_resources: List[MediaResource] = None):
        self.path = folder_path

        self.audio = Subfolder(folder_path, "00_audios")
        self.pdf = Subfolder(folder_path, "00_PDFs")
        self.image = Subfolder(folder_path, "00_images")
        self.infos = Subfolder(folder_path, "00_infos")
        self.rq = Subfolder(folder_path, "00_RQs")
        self.whisper = Subfolder(folder_path, "01_whisper")
        self.analyse = Subfolder(folder_path, "02_nlp")
        self.notes = Subfolder(folder_path, "03_notes")

        self.media_resources: List[MediaResource] = media_resources or []
        self.added_docs = set()
        # todo: if media_resources, maybe fill set?

        self.bag_of_words = BagOfWords()
        self.named_entities = NamedEntities()

        self.df = defaultdict(int)
        self.idf = defaultdict(int)
        self.sim_matrix = None
        
        if nlptools:
            self.process(nlptools)

    def get_image(self):
        candidates = []
        if self.image and self.image.path:
            candidates = os.listdir(self.image.path)
        if candidates:
            return os.path.join(self.image.path, candidates[0])

        for mr in self.media_resources:
            if mr.image:
                if os.path.exists(mr.image):
                    return mr.image
            if mr.info:
                if mr.info.thumbnail_urls:
                    # handle "if not mr.image"
                    file_path = mr.image
                    urllib.request.urlretrieve(
                        mr.info.thumbnail_urls[-1], file_path)

                    return file_path

    def calc_sim_matrix(self):
        # bag_sim_mat = similarity.compute_similarity_matrix(
        #     [doc.bag_of_words.get() for doc in self.documents], self.bag_of_words.get())

        bag_sim_mat = similarity.compute_similarity_matrix(
            [mr.nlp_analysis.bag_of_words.tf_idf for mr in self.media_resources], self.bag_of_words.get())

        # todo: option to eventually weigh named entities differently
        # nam_sim_mat = similarity.create_word_vectors(
        #     [doc.named_entities.get() for doc in self.documents], self.named_entities)
        self.sim_matrix = bag_sim_mat
        similarity.print_sim_matrix(self.sim_matrix, self.path)
        for mr in self.media_resources:
            should_have_nen = True
            if not mr.transcript:
                should_have_nen = False
            mr.nlp_analysis.save(should_have_nen=should_have_nen)

    def calc_rq_sim(self):
        if not self.rq.research_questions:
            return None

        def get_min_match(words: list[str], match_dict: dict[str, float]):
            found = []
            for word in words:
                if "-" in word:
                    value = 0
                    if word in mr.nlp_analysis.bag_of_words.tf:
                        value += mr.nlp_analysis.bag_of_words.tf[word]
                    synonyms = [
                        word.replace("-", " "),
                        word.replace("-", "")
                    ]
                    for syn in synonyms:
                        if syn in mr.nlp_analysis.bag_of_words.tf:
                            value += mr.nlp_analysis.bag_of_words.tf[syn]
                    splits = word.split("-")
                    value += get_min_match(splits, match_dict)
                    if value:
                        found.append(value)
                    else:
                        found = None
                        break
                else:
                    if word in mr.nlp_analysis.bag_of_words.tf:
                        found.append(
                            mr.nlp_analysis.bag_of_words.tf[word])
                    else:
                        found = None
                        break
            if found:
                return min(found)
            else:
                return 0

        relevant_keys = set()
        for rq in self.rq.get("rqs"):
            for group in rq.keywords.values():
                # todo: see
                relevant_keys.update(group)
                # for keyword in group:
                # relevant_keys.update(keyword.split(" "))
        reduced_research_questions = []
        for rq in self.rq.get("rqs"):
            res = {}
            ungrouped = {}
            for group in rq.keywords.values():
                ungrouped.update(group)
            for key in relevant_keys:
                value = 0
                if key in ungrouped:
                    # todo: utilize weight
                    value = ungrouped[key]
                    # value = 1
                res[key] = value
            reduced_research_questions.append(res)

        reduced_media_resources = []
        for mr in self.media_resources:
            res = {}
            for key in relevant_keys:
                # Todo: synonyms / related words
                key = key.lower()
                if key in mr.nlp_analysis.bag_of_words.tf:
                    res[key] = mr.nlp_analysis.bag_of_words.tf[key]
                else:
                    if " " in key:
                        subwords = key.split(" ")
                    else:
                        subwords = [key]
                    res[key] = get_min_match(
                        subwords, mr.nlp_analysis.bag_of_words.tf)
            reduced_media_resources.append(res)
        # TODO: calculate once per group, then again over all groups
        self.rq_sim_mat = similarity.compute_weighed_similarity(
            reduced_media_resources, reduced_research_questions)

        similarity.print_rq([mr.get("basename") for mr in self.media_resources], self.rq_sim_mat,
                            self.path, filename="RQ", additional_info=self.rq.research_questions)

    def add_media_resource(self, media_resource: MediaResource):
        if media_resource:
            self.media_resources.append(media_resource)

    def calc_idf(self):
        self.idf = {word: math.log(len(self.media_resources) / (frequency))
                    for word, frequency in self.df.items()}
        return self.idf

    def get_idf(self):
        if self.idf:
            return self.idf
        return self.calc_idf()

    def populate(self):
        for mr in self.media_resources:
            if mr.nlp_analysis.path not in self.added_docs:
                self.bag_of_words.add(mr.nlp_analysis.bag_of_words)
                self.named_entities.add(
                    mr.nlp_analysis.named_entities, mr.nlp_analysis.path)
                self.added_docs.add(mr.nlp_analysis.path)
                # document frequency
                for word in mr.nlp_analysis.bag_of_words.get().keys():
                    self.df[word] += 1

        for mr in self.media_resources:
            mr.nlp_analysis.bag_of_words.calc_tf_idf(self.get_idf())
        # self.bag_of_words.update_stops()
        self.named_entities.update()
        self.bag_of_words.sort()
        self.named_entities.sort()

    def get_dict(self):
        return {
            'bag_of_words': self.bag_of_words.words,
            'idf': self.idf,
            'named_entities': self.named_entities.get(),
            'stop_words': self.bag_of_words.stops,
            'documents': [os.path.basename(mr.nlp_analysis.path) for mr in self.media_resources]
        }

    def save(self, output_path=None, do_named_entities=True):
        if output_path is None:
            output_path = os.path.join(self.path, "rep_Folder")
        output_dict = self.get_dict()

        relevant = [len(self.media_resources), len(self.bag_of_words)]
        if do_named_entities:
            relevant.append(len(self.named_entities))

        if any(count == 0 for count in relevant):
            print("Error on Folder: " + output_path)
        dump_json(output_path, output_dict)

    def save_summary(self, output_path, do_named_entities=True):
        if not len(self.media_resources):
            print(f"No Documents found in folder: {output_path}")
            return None
        threshold = min(
            int(list(self.bag_of_words.get().values())[0] / 100), 10)
        threshold = max(threshold, 3)
        output_dict = {
            'bag_of_words': {k: v for k, v in list(self.bag_of_words.words.items())[:1000] if v >= threshold},
            'stop_words': {k: v for k, v in list(self.bag_of_words.stops.items())[:1000] if v >= threshold},
            'named_entities': {k: len(v) for k, v in list(self.named_entities.get().items())[:1000] if len(v) >= threshold},
            'documents': [os.path.basename(doc.nlp_analysis.path) for doc in self.media_resources]
        }

        relevant = [len(self.media_resources), len(self.bag_of_words)]
        if do_named_entities:
            relevant.append(len(self.named_entities))

        if any(count == 0 for count in relevant):
            print("Error on Folder: " + output_path)
        dump_json(output_path, output_dict)

    def process(self, nlptools=None):
        if nlptools is None:
            nlptools = NLPTools()
        audios = [os.path.join(self.audio.path, f)
                  for f in os.listdir(self.audio.path)]
        pdfs = self.pdf.BibResources.pdfs
        do_named_entities = False
        if audios:
            do_named_entities = True
            for idx, path in enumerate(audios):
                mr = MediaResource()
                mr.add_audio(path=path)
                mr.add_transcript(path=self.whisper.lookfor(path))
                mr.add_info(path=self.infos.lookfor(path))
                if not mr.transcript.text:
                    mr.transcript.text, mr.transcript.language = mr.info.get_transcript()
                mr.add_image(path=self.image.lookfor(path))
                # todo: eventually load thumbnails here
                mr.add_NLPFeatureAnalysis(
                    path=self.analyse.lookfor(path), nlptools=nlptools)

                mr.add_obsidian_note(path=self.notes.lookfor(path))
                self.add_media_resource(mr)
        if pdfs:
            for idx, path in enumerate(pdfs):
                mr = MediaResource()
                mr.add_pdf(path=path)
                mr.pdf.add_metadata(self.pdf.BibResources.get_metadata(path))
                mr.add_NLPFeatureAnalysis(path=self.analyse.lookfor(
                    path), nlptools=nlptools, do_named_entities=False)
                mr.add_obsidian_note(path=self.notes.lookfor(path))
                self.add_media_resource(mr)
                if idx % 100 == 99:
                    print(f"{idx+1} out of {len(pdfs)} added.")
            print(f"All {len(pdfs)} PDFs added.")

        if self.media_resources:
            self.populate()
            self.calc_sim_matrix()
            self.calc_rq_sim()
            self.save(os.path.join(self.path, "00_Folder.json"),
                      do_named_entities=do_named_entities)
            self.save_summary(os.path.join(
                self.path, "00_Folder_summary.json"), do_named_entities=do_named_entities)

    # Wordcloud
    def wordcloud(self):
        util_wordcloud.generate_wordcloud(
            self.bag_of_words.get(), os.path.join(self.path, "00_bag_of_words"))
        util_wordcloud.generate_wordcloud(
            self.named_entities.get_frequencies(), os.path.join(self.path, "00_named_entities"))
        found, not_found = self.image.lookfor("Folder")
        if found:
            mask = found[0]
            not_found = not_found[0]
        mask = util_wordcloud.generate_mask()
        util_wordcloud.generate_wordcloud_mask(
            self.bag_of_words.get(), mask, os.path.join(self.path, "00_bag_of_words_mask"))
        util_wordcloud.generate_wordcloud_mask(
            self.named_entities.get_frequencies(), mask, os.path.join(self.path, "00_named_entities"))


def main(input_path, output_path=None):
    if os.path.isfile(input_path):
        # process single file
        return MediaResource(paths={"transcript_file_path": input_path})
    elif os.path.isdir(input_path):
        # process all files in directory
        return Folder(input_path)
    else:
        print('Invalid input path.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Process JSON files and extract document features.')
    parser.add_argument(
        'input', help='path to JSON file or folder of JSON files')
    parser.add_argument(
        '-o', '--output', help='path to output file or folder (default: derived from input path)')
    args = parser.parse_args()

    main(args.input, args.output)



# Setup
import os
import sys
import importlib
import bnw_tools

bnw_tools_spec = importlib.util.find_spec('bnw_tools')
if bnw_tools_spec is None:
    # use development version
    sys.path.append(os.getcwd())
    import bnw_tools

from bnw_tools.publish.Obsidian import nlped_whispered_folder
from bnw_tools.publish import util_wordcloud
from bnw_tools.extract.nlp import util_nlp

folder_path = "D:/workspace/Zotero/SE2A-B4-2"
language = "en"

nlptools = util_nlp.NLPTools()

# NLP (Flair and SpaCy)
folder = util_nlp.Folder(
    folder_path, nlptools=nlptools, language=language)

# Wordcloud
if folder.media_resources:
    util_wordcloud.generate_wordcloud(
        folder.bag_of_words.get(), os.path.join(folder_path, "00_bag_of_words"))
    util_wordcloud.generate_wordcloud(
        folder.named_entities.get_frequencies(), os.path.join(folder_path, "00_named_entities"))
    mask_path = folder.get_image()
    if mask_path:
        mask = util_wordcloud.generate_mask(mask_path)
        util_wordcloud.generate_wordcloud_mask(
            folder.bag_of_words.get(), mask, os.path.join(folder_path, "00_bag_of_words_mask"))
        util_wordcloud.generate_wordcloud_mask(folder.named_entities.get_frequencies(
        ), mask, os.path.join(folder_path, "00_named_entities_mask"))

    # Obsidian
    nlped_whispered_folder.folder(folder, force=True)
